"""Abstractions for search functionality."""

from enum import Enum
from typing import Any, Optional
from uuid import UUID

from pydantic import BaseModel, Field

from .llm import GenerationConfig


class VectorSearchResult(BaseModel):
    """Result of a search operation."""

    fragment_id: UUID
    extraction_id: UUID
    document_id: UUID
    user_id: Optional[UUID]
    group_ids: list[UUID]
    score: float
    text: str
    metadata: dict[str, Any]

    def __str__(self) -> str:
        return f"VectorSearchResult(fragment_id={self.fragment_id}, extraction_id={self.extraction_id}, document_id={self.document_id}, score={self.score})"

    def __repr__(self) -> str:
        return self.__str__()

    def dict(self) -> dict:
        return {
            "fragment_id": self.fragment_id,
            "extraction_id": self.extraction_id,
            "document_id": self.document_id,
            "user_id": self.user_id,
            "group_ids": self.group_ids,
            "score": self.score,
            "text": self.text,
            "metadata": self.metadata,
        }

    class Config:
        json_schema_extra = {
            "fragment_id": "c68dc72e-fc23-5452-8f49-d7bd46088a96",
            "extraction_id": "3f3d47f3-8baf-58eb-8bc2-0171fb1c6e09",
            "document_id": "3e157b3a-8469-51db-90d9-52e7d896b49b",
            "user_id": "2acb499e-8428-543b-bd85-0d9098718220",
            "group_ids": [],
            "score": 0.23943702876567796,
            "text": "Example text from the document",
            "metadata": {
                "title": "example_document.pdf",
                "associatedQuery": "What is the capital of France?",
            },
        }


class KGLocalSearchResult(BaseModel):
    """Result of a local knowledge graph search operation."""

    query: str
    entities: dict[str, Any]
    relationships: dict[str, Any]
    communities: dict[str, Any]

    def __str__(self) -> str:
        return f"LocalSearchResult(query={self.query}, search_result={self.search_result})"

    def __repr__(self) -> str:
        return self.__str__()


class KGGlobalSearchResult(BaseModel):
    """Result of a global knowledge graph search operation."""

    query: str
    search_result: list[str]

    def __str__(self) -> str:
        return f"KGGlobalSearchResult(query={self.query}, search_result={self.search_result})"

    def __repr__(self) -> str:
        return self.__str__()

    def dict(self) -> dict:
        return {"query": self.query, "search_result": self.search_result}


class KGSearchResult(BaseModel):
    """Result of a knowledge graph search operation."""

    local_result: Optional[KGLocalSearchResult] = None
    global_result: Optional[KGGlobalSearchResult] = None

    def __str__(self) -> str:
        return f"KGSearchResult(local_result={self.local_result}, global_result={self.global_result})"

    def __repr__(self) -> str:
        return self.__str__()

    def dict(self) -> dict:
        return {
            "local_result": (
                self.local_result.dict() if self.local_result else None
            ),
            "global_result": (
                self.global_result.dict() if self.global_result else None
            ),
        }


class AggregateSearchResult(BaseModel):
    """Result of an aggregate search operation."""

    vector_search_results: Optional[list[VectorSearchResult]]
    kg_search_results: Optional[list[KGSearchResult]] = None

    def __str__(self) -> str:
        return f"AggregateSearchResult(vector_search_results={self.vector_search_results}, kg_search_results={self.kg_search_results})"

    def __repr__(self) -> str:
        return f"AggregateSearchResult(vector_search_results={self.vector_search_results}, kg_search_results={self.kg_search_results})"

    def dict(self) -> dict:
        return {
            "vector_search_results": (
                [result.dict() for result in self.vector_search_results]
                if self.vector_search_results
                else []
            ),
            "kg_search_results": self.kg_search_results or None,
        }


# TODO - stop duplication of this enum, move collections primitives to 'abstractions'
class IndexMeasure(str, Enum):
    """
    An enum representing the types of distance measures available for indexing.

    Attributes:
        cosine_distance (str): The cosine distance measure for indexing.
        l2_distance (str): The Euclidean (L2) distance measure for indexing.
        max_inner_product (str): The maximum inner product measure for indexing.
    """

    cosine_distance = "cosine_distance"
    l2_distance = "l2_distance"
    max_inner_product = "max_inner_product"


class HybridSearchSettings(BaseModel):
    full_text_weight: float = Field(
        default=1.0, description="Weight to apply to full text search"
    )
    semantic_weight: float = Field(
        default=5.0, description="Weight to apply to semantic search"
    )
    full_text_limit: int = Field(
        default=200,
        description="Maximum number of results to return from full text search",
    )
    rrf_k: int = Field(
        default=50, description="K-value for RRF (Rank Reciprocal Fusion)"
    )


class VectorSearchSettings(BaseModel):
    use_vector_search: bool = Field(
        default=True, description="Whether to use vector search"
    )
    use_hybrid_search: bool = Field(
        default=False,
        description="Whether to perform a hybrid search (combining vector and keyword search)",
    )
    filters: dict[str, Any] = Field(
        default_factory=dict,
        description="Filters to apply to the vector search",
    )
    search_limit: int = Field(
        default=10,
        description="Maximum number of results to return",
        ge=1,
        le=1_000,
    )
    selected_group_ids: list[UUID] = Field(
        default_factory=list,
        description="Group IDs to search for",
    )
    index_measure: IndexMeasure = Field(
        default=IndexMeasure.cosine_distance,
        description="The distance measure to use for indexing",
    )
    include_values: bool = Field(
        default=True,
        description="Whether to include search score values in the search results",
    )
    include_metadatas: bool = Field(
        default=True,
        description="Whether to include element metadata in the search results",
    )
    probes: Optional[int] = Field(
        default=10,
        description="Number of ivfflat index lists to query. Higher increases accuracy but decreases speed.",
    )
    ef_search: Optional[int] = Field(
        default=40,
        description="Size of the dynamic candidate list for HNSW index search. Higher increases accuracy but decreases speed.",
    )
    hybrid_search_settings: Optional[HybridSearchSettings] = Field(
        default=HybridSearchSettings(),
        description="Settings for hybrid search",
    )

    class Config:
        json_encoders = {UUID: str}
        json_schema_extra = {
            "use_vector_search": True,
            "use_hybrid_search": True,
            "filters": {"category": "technology"},
            "search_limit": 20,
            "selected_group_ids": [
                "2acb499e-8428-543b-bd85-0d9098718220",
                "3e157b3a-8469-51db-90d9-52e7d896b49b",
            ],
            "index_measure": "cosine_distance",
            "include_metadata": True,
            "probes": 10,
            "ef_search": 40,
            "hybrid_search_settings": {
                "full_text_weight": 1.0,
                "semantic_weight": 5.0,
                "full_text_limit": 200,
                "rrf_k": 50,
            },
        }

    def model_dump(self, *args, **kwargs):
        dump = super().model_dump(*args, **kwargs)
        dump["selected_group_ids"] = [
            str(uuid) for uuid in dump["selected_group_ids"]
        ]
        return dump


class KGSearchSettings(BaseModel):
    use_kg_search: bool = False
    kg_search_type: str = "global"  # 'global' or 'local'
    kg_search_level: Optional[str] = None
    kg_search_generation_config: Optional[GenerationConfig] = Field(
        default_factory=GenerationConfig
    )
    entity_types: list = []
    relationships: list = []
    max_community_description_length: int = 4096 * 4
    max_llm_queries_for_global_search: int = 250
    local_search_limits: dict[str, int] = {
        "__Entity__": 20,
        "__Relationship__": 20,
        "__Community__": 20,
    }

    class Config:
        json_encoders = {UUID: str}
        json_schema_extra = {
            "use_kg_search": True,
            "kg_search_type": "global",
            "kg_search_level": "global",
            "kg_search_generation_config": GenerationConfig.Config.json_schema_extra,
            "entity_types": ["Person", "Organization"],
            "relationships": ["founder", "CEO"],
            "max_community_description_length": 4096 * 4,
            "max_llm_queries_for_global_search": 250,
            "local_search_limits": {
                "__Entity__": 20,
                "__Relationship__": 20,
                "__Community__": 20,
            },
        }



import fire

from r2r import R2RBuilder, SerperClient, WebSearchPipe


def run_rag_pipeline():
    # Create search pipe override and pipes
    web_search_pipe = WebSearchPipe(
        serper_client=SerperClient()  # TODO - Develop a `WebSearchProvider` for configurability
    )

    app = R2RBuilder().with_vector_search_pipe(web_search_pipe).build()

    app.serve()


if __name__ == "__main__":
    fire.Fire(run_rag_pipeline)



"""Abstraction for a vector that can be stored in the system."""

from enum import Enum
from typing import Any
from uuid import UUID


class VectorType(Enum):
    FIXED = "FIXED"


class Vector:
    """A vector with the option to fix the number of elements."""

    def __init__(
        self,
        data: list[float],
        type: VectorType = VectorType.FIXED,
        length: int = -1,
    ):
        self.data = data
        self.type = type
        self.length = length

        if (
            self.type == VectorType.FIXED
            and length > 0
            and len(data) != length
        ):
            raise ValueError(f"Vector must be exactly {length} elements long.")

    def __repr__(self) -> str:
        return (
            f"Vector(data={self.data}, type={self.type}, length={self.length})"
        )


class VectorEntry:
    """A vector entry that can be stored directly in supported vector databases."""

    def __init__(
        self,
        fragment_id: UUID,
        extraction_id: UUID,
        document_id: UUID,
        user_id: UUID,
        group_ids: list[UUID],
        vector: Vector,
        text: str,
        metadata: dict[str, Any],
    ):
        """Create a new VectorEntry object."""
        self.fragment_id = fragment_id
        self.extraction_id = extraction_id
        self.document_id = document_id
        self.user_id = user_id
        self.group_ids = group_ids
        self.vector = vector
        self.text = text
        self.metadata = metadata

    def __str__(self) -> str:
        """Return a string representation of the VectorEntry."""
        return (
            f"VectorEntry(fragment_id={self.fragment_id}, "
            f"extraction_id={self.extraction_id}, "
            f"document_id={self.document_id}, "
            f"user_id={self.user_id}, "
            f"group_ids={self.group_ids}, "
            f"vector={self.vector}, "
            f"text={self.text}, "
            f"metadata={self.metadata})"
        )

    def __repr__(self) -> str:
        """Return an unambiguous string representation of the VectorEntry."""
        return self.__str__()



from cli.command_group import cli
from cli.commands import (
    auth,
    ingestion,
    management,
    restructure,
    retrieval,
    server,
)

# Auth
cli.add_command(auth.generate_private_key)

# Ingesetion
cli.add_command(ingestion.ingest_files)
cli.add_command(ingestion.update_files)
cli.add_command(ingestion.ingest_sample_file)
cli.add_command(ingestion.ingest_sample_files)

# Management
cli.add_command(management.analytics)
cli.add_command(management.app_settings)
cli.add_command(management.users_overview)
cli.add_command(management.documents_overview)
cli.add_command(management.document_chunks)
cli.add_command(management.inspect_knowledge_graph)

# Restructure
cli.add_command(restructure.enrich_graph)

# Retrieval
cli.add_command(retrieval.search)
cli.add_command(retrieval.rag)

# Server
cli.add_command(server.health)
cli.add_command(server.server_stats)
cli.add_command(server.logs)
cli.add_command(server.docker_down)
cli.add_command(server.generate_report)
cli.add_command(server.serve)
cli.add_command(server.update)
cli.add_command(server.version)


def main():
    cli()


if __name__ == "__main__":
    main()



from datetime import datetime
from typing import Optional
from uuid import UUID

from sqlalchemy import text

from core.base.abstractions import R2RException, UserStats
from core.base.api.models.auth.responses import UserResponse
from core.base.utils import generate_id_from_label

from .base import DatabaseMixin, QueryBuilder


class UserMixin(DatabaseMixin):
    def create_table(self):
        query = f"""
        CREATE TABLE IF NOT EXISTS {self._get_table_name('users')} (
            user_id UUID PRIMARY KEY DEFAULT uuid_generate_v4(),
            email TEXT UNIQUE NOT NULL,
            hashed_password TEXT NOT NULL,
            is_superuser BOOLEAN DEFAULT FALSE,
            is_active BOOLEAN DEFAULT TRUE,
            is_verified BOOLEAN DEFAULT FALSE,
            verification_code TEXT,
            verification_code_expiry TIMESTAMPTZ,
            name TEXT,
            bio TEXT,
            profile_picture TEXT,
            reset_token TEXT,
            reset_token_expiry TIMESTAMPTZ,
            group_ids UUID[] NULL,
            created_at TIMESTAMPTZ DEFAULT NOW(),
            updated_at TIMESTAMPTZ DEFAULT NOW()
        );
        """
        self.execute_query(query)

    def get_user_by_id(self, user_id: UUID) -> Optional[UserResponse]:
        query, params = (
            QueryBuilder(self._get_table_name("users"))
            .select(
                [
                    "user_id",
                    "email",
                    "hashed_password",
                    "is_superuser",
                    "is_active",
                    "is_verified",
                    "created_at",
                    "updated_at",
                    "name",
                    "profile_picture",
                    "bio",
                    "group_ids",
                ]
            )
            .where("user_id = :user_id", user_id=user_id)
            .build()
        )
        result = self.execute_query(query, params).fetchone()
        return (
            UserResponse(
                id=result[0],
                email=result[1],
                hashed_password=result[2],
                is_superuser=result[3],
                is_active=result[4],
                is_verified=result[5],
                created_at=result[6],
                updated_at=result[7],
                name=result[8],
                profile_picture=result[9],
                bio=result[10],
                group_ids=result[11],
            )
            if result
            else None
        )

    def get_user_by_email(self, email: str) -> UserResponse:
        query, params = (
            QueryBuilder(self._get_table_name("users"))
            .select(
                [
                    "user_id",
                    "email",
                    "hashed_password",
                    "is_superuser",
                    "is_active",
                    "is_verified",
                    "created_at",
                    "updated_at",
                    "name",
                    "profile_picture",
                    "bio",
                    "group_ids",
                ]
            )
            .where("email = :email", email=email)
            .build()
        )
        result = self.execute_query(query, params).fetchone()
        if not result:
            raise R2RException(status_code=404, message="User not found")

        return UserResponse(
            id=result[0],
            email=result[1],
            hashed_password=result[2],
            is_superuser=result[3],
            is_active=result[4],
            is_verified=result[5],
            created_at=result[6],
            updated_at=result[7],
            name=result[8],
            profile_picture=result[9],
            bio=result[10],
            group_ids=result[11],
        )

    def create_user(self, email: str, password: str) -> UserResponse:
        try:
            if self.get_user_by_email(email):
                raise R2RException(
                    status_code=400,
                    message="User with this email already exists",
                )
        except R2RException as e:
            if e.status_code != 404:
                raise e

        hashed_password = self.crypto_provider.get_password_hash(password)
        query = f"""
            INSERT INTO {self._get_table_name('users')}
            (email, user_id, hashed_password, group_ids)
            VALUES (:email, :user_id, :hashed_password, :group_ids)
            RETURNING user_id, email, is_superuser, is_active, is_verified, created_at, updated_at, group_ids
        """
        params = {
            "email": email,
            "user_id": generate_id_from_label(email),
            "hashed_password": hashed_password,
            "group_ids": [],
        }
        result = self.execute_query(query, params).fetchone()
        if not result:
            raise R2RException(
                status_code=500, message="Failed to create user"
            )

        return UserResponse(
            id=result[0],
            email=result[1],
            is_superuser=result[2],
            is_active=result[3],
            is_verified=result[4],
            created_at=result[5],
            updated_at=result[6],
            group_ids=result[7],
            hashed_password=hashed_password,
        )

    def update_user(self, user: UserResponse) -> UserResponse:
        if not self.get_user_by_id(user.id):
            raise R2RException(status_code=404, message="User not found")

        query = f"""
            UPDATE {self._get_table_name('users')}
            SET email = :email, is_superuser = :is_superuser, is_active = :is_active,
                is_verified = :is_verified, updated_at = NOW(), name = :name,
                profile_picture = :profile_picture, bio = :bio, group_ids = :group_ids
            WHERE user_id = :id
            RETURNING user_id, email, is_superuser, is_active, is_verified, created_at,
                      updated_at, name, profile_picture, bio, group_ids
        """
        result = self.execute_query(
            query, user.dict(exclude={"hashed_password"})
        ).fetchone()
        if not result:
            raise R2RException(
                status_code=500, message="Failed to update user"
            )
        return UserResponse(
            id=result[0],
            email=result[1],
            is_superuser=result[2],
            is_active=result[3],
            is_verified=result[4],
            created_at=result[5],
            updated_at=result[6],
            name=result[7],
            profile_picture=result[8],
            bio=result[9],
            group_ids=result[10],
        )

    def delete_user(self, user_id: UUID) -> None:
        # Get the groups the user belongs to
        group_query = f"""
            SELECT group_ids FROM {self._get_table_name('users')}
            WHERE user_id = :user_id
        """
        group_result = self.execute_query(
            group_query, {"user_id": user_id}
        ).fetchone()

        if not group_result:
            raise R2RException(status_code=404, message="User not found")

        user_groups = group_result[0]

        # Remove user from documents
        doc_update_query = f"""
            UPDATE {self._get_table_name('document_info')}
            SET user_id = NULL
            WHERE user_id = :user_id
        """
        self.execute_query(doc_update_query, {"user_id": user_id})

        # Delete the user
        delete_query = f"""
            DELETE FROM {self._get_table_name('users')}
            WHERE user_id = :user_id
            RETURNING user_id
        """

        result = self.execute_query(
            delete_query, {"user_id": user_id}
        ).fetchone()

        if not result:
            raise R2RException(status_code=404, message="User not found")

    def update_user_password(self, user_id: UUID, new_hashed_password: str):
        query = f"""
            UPDATE {self._get_table_name('users')}
            SET hashed_password = :new_hashed_password, updated_at = NOW()
            WHERE user_id = :user_id
        """
        self.execute_query(
            query,
            {"user_id": user_id, "new_hashed_password": new_hashed_password},
        )

    def get_all_users(self) -> list[UserResponse]:
        query = f"""
            SELECT user_id, email, is_superuser, is_active, is_verified, created_at, updated_at, group_ids
            FROM {self._get_table_name('users')}
        """
        results = self.execute_query(query).fetchall()
        return [
            UserResponse(
                id=row[0],
                email=row[1],
                hashed_password="null",
                is_superuser=row[2],
                is_active=row[3],
                is_verified=row[4],
                created_at=row[5],
                updated_at=row[6],
                group_ids=row[7],
            )
            for row in results
        ]

    def store_verification_code(
        self, user_id: UUID, verification_code: str, expiry: datetime
    ):
        query = f"""
            UPDATE {self._get_table_name('users')}
            SET verification_code = :code, verification_code_expiry = :expiry
            WHERE user_id = :user_id
        """
        self.execute_query(
            query,
            {"code": verification_code, "expiry": expiry, "user_id": user_id},
        )

    def verify_user(self, verification_code: str) -> None:
        query = f"""
            UPDATE {self._get_table_name('users')}
            SET is_verified = TRUE, verification_code = NULL, verification_code_expiry = NULL
            WHERE verification_code = :code AND verification_code_expiry > NOW()
            RETURNING user_id
        """
        result = self.execute_query(
            query, {"code": verification_code}
        ).fetchone()
        if not result:
            raise R2RException(
                status_code=400, message="Invalid or expired verification code"
            )
        return None

    def remove_verification_code(self, verification_code: str):
        query = f"""
            UPDATE {self._get_table_name('users')}
            SET verification_code = NULL, verification_code_expiry = NULL
            WHERE verification_code = :code
        """
        self.execute_query(query, {"code": verification_code})

    def expire_verification_code(self, user_id: UUID):
        query = f"""
            UPDATE {self._get_table_name('users')}
            SET verification_code_expiry = NOW() - INTERVAL '1 day'
            WHERE user_id = :user_id
        """
        self.execute_query(query, {"user_id": user_id})

    def store_reset_token(
        self, user_id: UUID, reset_token: str, expiry: datetime
    ):
        query = f"""
            UPDATE {self._get_table_name('users')}
            SET reset_token = :token, reset_token_expiry = :expiry
            WHERE user_id = :user_id
        """
        self.execute_query(
            query, {"token": reset_token, "expiry": expiry, "user_id": user_id}
        )

    def get_user_id_by_reset_token(self, reset_token: str) -> Optional[UUID]:
        query = f"""
            SELECT user_id FROM {self._get_table_name('users')}
            WHERE reset_token = :token AND reset_token_expiry > NOW()
        """
        result = self.execute_query(query, {"token": reset_token}).fetchone()
        return result[0] if result else None

    def remove_reset_token(self, user_id: UUID):
        query = f"""
            UPDATE {self._get_table_name('users')}
            SET reset_token = NULL, reset_token_expiry = NULL
            WHERE user_id = :user_id
        """
        self.execute_query(query, {"user_id": user_id})

    def remove_user_from_all_groups(self, user_id: UUID):
        query = f"""
            UPDATE {self._get_table_name('users')}
            SET group_ids = ARRAY[]::UUID[]
            WHERE user_id = :user_id
        """
        self.execute_query(query, {"user_id": user_id})

    def add_user_to_group(self, user_id: UUID, group_id: UUID) -> None:
        if not self.get_user_by_id(user_id):
            raise R2RException(status_code=404, message="User not found")

        query = f"""
            UPDATE {self._get_table_name('users')}
            SET group_ids = array_append(group_ids, :group_id)
            WHERE user_id = :user_id AND NOT (:group_id = ANY(group_ids))
            RETURNING user_id
        """
        result = self.execute_query(
            query, {"user_id": user_id, "group_id": group_id}
        ).fetchone()
        if not result:
            raise R2RException(
                status_code=400, message="User already in group"
            )
        return None

    def remove_user_from_group(self, user_id: UUID, group_id: UUID) -> None:
        if not self.get_user_by_id(user_id):
            raise R2RException(status_code=404, message="User not found")

        query = f"""
            UPDATE {self._get_table_name('users')}
            SET group_ids = array_remove(group_ids, :group_id)
            WHERE user_id = :user_id AND :group_id = ANY(group_ids)
            RETURNING user_id
        """
        result = self.execute_query(
            query, {"user_id": user_id, "group_id": group_id}
        ).fetchone()
        if not result:
            raise R2RException(
                status_code=400,
                message="User is not a member of the specified group",
            )
        return None

    def mark_user_as_superuser(self, user_id: UUID):
        query = text(
            f"""
        UPDATE users_{self.collection_name}
        SET is_superuser = TRUE, is_verified = TRUE, verification_code = NULL, verification_code_expiry = NULL
        WHERE user_id = :user_id
        """
        )

        with self.vx.Session() as sess:
            sess.execute(query, {"user_id": user_id})
            sess.commit()

    def get_users_in_group(
        self, group_id: UUID, offset: int = 0, limit: int = 100
    ) -> list[UserResponse]:
        query = f"""
            SELECT user_id, email, is_superuser, is_active, is_verified, created_at, updated_at, name, profile_picture, bio, group_ids
            FROM {self._get_table_name('users')}
            WHERE :group_id = ANY(group_ids)
            ORDER BY email
            OFFSET :offset
            LIMIT :limit
        """
        results = self.execute_query(
            query, {"group_id": group_id, "offset": offset, "limit": limit}
        ).fetchall()
        return [
            UserResponse(
                id=row[0],
                email=row[1],
                is_superuser=row[2],
                is_active=row[3],
                is_verified=row[4],
                created_at=row[5],
                updated_at=row[6],
                name=row[7],
                profile_picture=row[8],
                bio=row[9],
                group_ids=row[10],
            )
            for row in results
        ]

    def get_user_id_by_verification_code(
        self, verification_code: str
    ) -> Optional[UUID]:
        query = text(
            f"""
        SELECT user_id FROM users_{self.collection_name}
        WHERE verification_code = :code AND verification_code_expiry > NOW()
        """
        )

        with self.vx.Session() as sess:
            result = sess.execute(query, {"code": verification_code})
            user_data = result.fetchone()

        if not user_data:
            raise R2RException(
                status_code=400, message="Invalid or expired verification code"
            )
        return user_data[0]

    def mark_user_as_verified(self, user_id: UUID):
        query = text(
            f"""
        UPDATE users_{self.collection_name}
        SET is_verified = TRUE, verification_code = NULL, verification_code_expiry = NULL
        WHERE user_id = :user_id
        """
        )

        with self.vx.Session() as sess:
            result = sess.execute(query, {"user_id": user_id})
            sess.commit()

        if not result.rowcount:
            raise R2RException(status_code=404, message="User not found")

    def get_users_overview(
        self,
        user_ids: Optional[list[UUID]] = None,
        offset: int = 0,
        limit: int = 100,
    ) -> list[UserStats]:
        query = f"""
            WITH user_docs AS (
                SELECT
                    u.user_id,
                    u.email,
                    u.is_superuser,
                    u.is_active,
                    u.is_verified,
                    u.created_at,
                    u.updated_at,
                    u.group_ids,
                    COUNT(d.document_id) AS num_files,
                    COALESCE(SUM(d.size_in_bytes), 0) AS total_size_in_bytes,
                    ARRAY_AGG(d.document_id) FILTER (WHERE d.document_id IS NOT NULL) AS document_ids
                FROM {self._get_table_name('users')} u
                LEFT JOIN {self._get_table_name('document_info')} d ON u.user_id = d.user_id
                {f"WHERE u.user_id = ANY(CAST(:user_ids AS UUID[]))" if user_ids else ""}
                GROUP BY u.user_id, u.email, u.is_superuser, u.is_active, u.is_verified, u.created_at, u.updated_at, u.group_ids
            )
            SELECT *
            FROM user_docs
            ORDER BY email
            OFFSET :offset
            LIMIT :limit
        """

        params = {"offset": offset, "limit": limit}
        if user_ids:
            params["user_ids"] = [str(ele) for ele in user_ids]

        results = self.execute_query(query, params).fetchall()

        return [
            UserStats(
                user_id=row[0],
                email=row[1],
                is_superuser=row[2],
                is_active=row[3],
                is_verified=row[4],
                created_at=row[5],
                updated_at=row[6],
                group_ids=row[7] or [],
                num_files=row[8],
                total_size_in_bytes=row[9],
                document_ids=row[10] or [],
            )
            for row in results
        ]



from typing import AsyncGenerator, Union

from core.base import ChunkingProvider, Method
from core.base.abstractions.document import DocumentExtraction


class UnstructuredChunkingProvider(ChunkingProvider):
    def __init__(self, config):
        try:
            from unstructured.chunking.basic import chunk_elements

            self.chunk_elements = chunk_elements
            from unstructured.chunking.title import chunk_by_title

            self.chunk_by_title = chunk_by_title

            from unstructured.documents.elements import Text

            self.Text = Text
        except ImportError:
            raise ImportError(
                "Please install the unstructured package to use the unstructured chunking provider."
            )
        super().__init__(config)

    async def chunk(
        self, parsed_document: Union[str, DocumentExtraction]
    ) -> AsyncGenerator[str, None]:

        # as unstructured has already partitioned the document, we can yield the text directly
        if parsed_document.metadata.get("partitioned_by_unstructured", False):
            yield parsed_document.data

        else:
            if self.config.method == Method.BY_TITLE:
                chunks = self.chunk_by_title(
                    [self.Text(text=parsed_document.data)],
                    max_characters=self.config.chunk_size,
                    new_after_n_chars=self.config.max_chunk_size
                    or self.config.chunk_size,
                    overlap=self.config.chunk_overlap,
                )
            else:
                chunks = self.chunk_elements(
                    [self.Text(text=parsed_document.data)],
                    max_characters=self.config.chunk_size,
                    new_after_n_chars=self.config.max_chunk_size
                    or self.config.chunk_size,
                    overlap=self.config.chunk_overlap,
                )
            for chunk in chunks:
                yield chunk.text



from typing import Optional


class ServerMethods:
    @staticmethod
    async def health(self) -> dict:
        return await self._make_request("GET", "health")

    @staticmethod
    async def server_stats(client) -> dict:
        """
        Get statistics about the server, including the start time, uptime, CPU usage, and memory usage.

        Returns:
            dict: The server statistics.
        """
        return await client._make_request("GET", "server_stats")

    @staticmethod
    async def logs(
        client,
        run_type_filter: Optional[str] = None,
        max_runs: int = None,
    ) -> dict:
        """
        Get logs from the server.

        Args:
            run_type_filter (Optional[str]): The run type to filter by.
            max_runs (int): Specifies the maximum number of runs to return. Values outside the range of 1 to 1000 will be adjusted to the nearest valid value with a default of 100.

        Returns:
            dict: The logs from the server.
        """
        params = {}
        if run_type_filter is not None:
            params["run_type_filter"] = run_type_filter
        if max_runs is not None:
            params["max_runs"] = max_runs
        return await client._make_request("GET", "logs", params=params)



from .models import KGEnrichmentResponse, KGEnrichmentSettings


class RestructureMethods:
    @staticmethod
    async def enrich_graph(
        client,
        KGEnrichmentSettings: KGEnrichmentSettings = KGEnrichmentSettings(),
    ) -> KGEnrichmentResponse:
        """
        Perform graph enrichment over the entire graph.

        Args:
            KGEnrichmentSettings (KGEnrichmentSettings): Settings for the graph enrichment process.

        Returns:
            KGEnrichmentResponse: Results of the graph enrichment process.
        """
        if not isinstance(KGEnrichmentSettings, dict):
            KGEnrichmentSettings = KGEnrichmentSettings.model_dump()

        data = {
            "KGEnrichmentSettings": KGEnrichmentSettings,
        }
        return await client._make_request("POST", "enrich_graph", json=data)



from typing import Union

from fastapi import Body, Depends

from core.base import KGEnrichmentSettings
from core.main.api.routes.base_router import BaseRouter, RunType
from core.main.engine import R2REngine


class RestructureRouter(BaseRouter):
    def __init__(
        self, engine: R2REngine, run_type: RunType = RunType.RESTRUCTURE
    ):
        super().__init__(engine, run_type)
        self.setup_routes()

    def setup_routes(self):
        @self.router.post("/enrich_graph")
        @self.base_endpoint
        async def enrich_graph(
            KGEnrichmentSettings: Union[dict, KGEnrichmentSettings] = Body(
                ...,
                description="Settings for knowledge graph enrichment",
            ),
            auth_user=(
                Depends(self.engine.providers.auth.auth_wrapper)
                if self.engine.providers.auth
                else None
            ),
        ):
            """
            Perform graph enrichment, e.g. GraphRAG, over the ingested documents.

            Returns:
                Dict[str, Any]: Results of the graph enrichment process.
            """
            return await self.engine.aenrich_graph()



import logging
import os
import time
from io import BytesIO
from typing import AsyncGenerator

from core.base import (
    Document,
    DocumentExtraction,
    DocumentType,
    ParsingProvider,
    generate_id_from_label,
)

logger = logging.getLogger(__name__)


class UnstructuredParsingProvider(ParsingProvider):
    def __init__(self, use_api, config):
        if config.excluded_parsers:
            logger.warning(
                "Excluded parsers are not supported by the unstructured parsing provider."
            )

        self.use_api = use_api
        if self.use_api:
            try:
                from unstructured_client import UnstructuredClient
                from unstructured_client.models import operations, shared
                from unstructured_client.models.errors import SDKError

            except ImportError:
                raise ImportError(
                    "Please install the unstructured package to use the unstructured parsing provider."
                )

            try:
                self.unstructured_api_auth = os.environ["UNSTRUCTURED_API_KEY"]
            except KeyError:
                raise ValueError(
                    "UNSTRUCTURED_API_KEY environment variable is not set"
                )

            self.unstructured_api_url = os.environ.get(
                "UNSTRUCTURED_API_URL",
                "https://api.unstructured.io/general/v0/general",
            )

            self.client = UnstructuredClient(
                api_key_auth=self.unstructured_api_auth,
                server_url=self.unstructured_api_url,
            )
            self.shared = shared
            self.operations = operations

        else:
            try:
                from unstructured.partition.auto import partition

                self.partition = partition

            except ImportError:
                raise ImportError(
                    "Please install the unstructured package to use the unstructured parsing provider."
                )

        super().__init__(config)

    async def parse(
        self, document: Document
    ) -> AsyncGenerator[DocumentExtraction, None]:
        data = document.data
        if isinstance(data, bytes):
            data = BytesIO(data)

        # TODO - Include check on excluded parsers here.
        t0 = time.time()
        if self.use_api:
            logger.info(f"Using API to parse document {document.id}")
            files = self.shared.Files(
                content=data.read() if isinstance(data, BytesIO) else data,
                file_name=document.metadata.get("filename", "unknown_file"),
            )

            req = self.operations.PartitionRequest(
                self.shared.PartitionParameters(
                    files=files,
                    split_pdf_page=True,
                    split_pdf_allow_failed=True,
                    split_pdf_concurrency_level=15,
                )
            )
            elements = self.client.general.partition(req)
            elements = list(elements.elements)

        else:
            logger.info(
                f"Using local unstructured to parse document {document.id}"
            )
            elements = self.partition(
                file=data, **self.config.chunking_config.dict()
            )

        for iteration, element in enumerate(elements):

            for key, value in element.items():
                if key != "text":
                    if key == "metadata":
                        for k, v in value.items():
                            if k not in document.metadata:
                                document.metadata[k] = v
                            else:
                                document.metadata[f"unstructured_{k}"] = v
                    elif key in document.metadata:
                        document.metadata[f"unstructured_{key}"] = value
                    else:
                        document.metadata[key] = value
                else:
                    text = value

            # indicate that the document was chunked using unstructured
            # nullifies the need for chunking in the pipeline
            document.metadata["partitioned_by_unstructured"] = True

            # creating the text extraction
            extraction = DocumentExtraction(
                id=generate_id_from_label(f"{document.id}-{iteration}"),
                document_id=document.id,
                user_id=document.user_id,
                group_ids=document.group_ids,
                data=text,
                metadata=document.metadata,
            )

            yield extraction

        logger.debug(
            f"Parsed document with id={document.id}, title={document.metadata.get('title', None)}, "
            f"user_id={document.metadata.get('user_id', None)}, metadata={document.metadata} "
            f"into {iteration + 1} extractions in t={time.time() - t0:.2f} seconds."
        )

    def get_parser_for_document_type(self, doc_type: DocumentType) -> str:
        return "unstructured"



from .routes.auth.base import AuthRouter
from .routes.base_router import BaseRouter
from .routes.ingestion.base import IngestionRouter
from .routes.management.base import ManagementRouter
from .routes.restructure.base import RestructureRouter
from .routes.retrieval.base import RetrievalRouter

__all__ = [
    # Routes
    "AuthRouter",
    "IngestionRouter",
    "ManagementRouter",
    "RetrievalRouter",
    "BaseRouter",
    "RestructureRouter",
]



import logging
from typing import Any, Dict

from core.base import R2RException, RunLoggingSingleton, RunManager
from core.base.abstractions import KGEnrichmentSettings

from ..abstractions import R2RAgents, R2RPipelines, R2RProviders
from ..assembly.config import R2RConfig
from .base import Service

logger = logging.getLogger(__name__)


class RestructureService(Service):
    def __init__(
        self,
        config: R2RConfig,
        providers: R2RProviders,
        pipelines: R2RPipelines,
        agents: R2RAgents,
        run_manager: RunManager,
        logging_connection: RunLoggingSingleton,
    ):

        super().__init__(
            config,
            providers,
            pipelines,
            agents,
            run_manager,
            logging_connection,
        )

    async def enrich_graph(
        self,
        enrich_graph_settings: KGEnrichmentSettings = KGEnrichmentSettings(),
    ) -> Dict[str, Any]:
        """
        Perform graph enrichment.

        Returns:
            Dict[str, Any]: Results of the graph enrichment process.
        """
        try:
            # Assuming there's a graph enrichment pipeline
            return await self.pipelines.kg_enrichment_pipeline.run(
                input=[],
                run_manager=self.run_manager,
            )

        except Exception as e:
            logger.error(f"Error during graph enrichment: {str(e)}")
            raise R2RException(
                status_code=500, message=f"Graph enrichment failed: {str(e)}"
            )



from abc import ABC, abstractmethod
from enum import Enum
from typing import AsyncGenerator, Optional, Union

from ..abstractions.document import DocumentExtraction
from .base import Provider, ProviderConfig


class Method(str, Enum):
    BY_TITLE = "by_title"
    BASIC = "basic"
    RECURSIVE = "recursive"


class ChunkingConfig(ProviderConfig):
    provider: str = "r2r"
    method: Method = Method.RECURSIVE
    chunk_size: int = 512
    chunk_overlap: int = 20
    max_chunk_size: Optional[int] = None

    def validate(self) -> None:
        if self.provider not in self.supported_providers:
            raise ValueError(f"Provider {self.provider} is not supported.")
        if self.chunk_size <= 0:
            raise ValueError("chunk_size must be greater than 0")
        if self.chunk_overlap < 0:
            raise ValueError("chunk_overlap must be non-negative")

    @property
    def supported_providers(self) -> list[str]:
        return ["r2r", "unstructured", None]

    class Config:
        json_schema_extra = {
            "type": "object",
            "properties": {
                "provider": {"type": "string"},
                "method": {"type": "string"},
                "chunk_size": {"type": "integer"},
                "chunk_overlap": {"type": "integer"},
                "max_chunk_size": {"type": "integer"},
            },
            "required": ["provider", "method", "chunk_size", "chunk_overlap"],
            "example": {
                "provider": "r2r",
                "method": "recursive",
                "chunk_size": 512,
                "chunk_overlap": 20,
                "max_chunk_size": 1024,
            },
        }


class ChunkingProvider(Provider, ABC):
    def __init__(self, config: ChunkingConfig):
        super().__init__(config)
        self.config = config

    @abstractmethod
    async def chunk(
        self, parsed_document: Union[str, DocumentExtraction]
    ) -> AsyncGenerator[str, None]:
        """Chunk the parsed document using the configured chunking strategy."""
        pass



import click

from cli.command_group import cli
from cli.utils.timer import timer


@cli.command()
@click.pass_obj
def enrich_graph(client):
    """
    Perform graph enrichment over the entire graph.
    """
    with timer():
        response = client.enrich_graph()

    click.echo(response)



# TODO - Cleanup the handling for non-auth configurations
import json
from datetime import datetime, timezone
from typing import Optional
from uuid import UUID

import psutil
from fastapi import Body, Depends, Path, Query
from pydantic import Json

from core.base import R2RException
from core.base.api.models.management.responses import (
    WrappedAddUserResponse,
    WrappedAnalyticsResponse,
    WrappedAppSettingsResponse,
    WrappedDocumentChunkResponse,
    WrappedDocumentOverviewResponse,
    WrappedGroupListResponse,
    WrappedGroupOverviewResponse,
    WrappedGroupResponse,
    WrappedKnowledgeGraphResponse,
    WrappedLogResponse,
    WrappedPromptResponse,
    WrappedScoreCompletionResponse,
    WrappedServerStatsResponse,
    WrappedUserOverviewResponse,
)
from core.base.logging import AnalysisTypes, LogFilterCriteria

from ....engine import R2REngine
from ..base_router import BaseRouter, RunType


class ManagementRouter(BaseRouter):
    def __init__(
        self, engine: R2REngine, run_type: RunType = RunType.MANAGEMENT
    ):
        super().__init__(engine, run_type)
        self.start_time = datetime.now(timezone.utc)
        self.setup_routes()

    # TODO: remove this from the management route, it should be at the base of the server
    def setup_routes(self):
        @self.router.get("/health")
        @self.base_endpoint
        async def health_check():
            return {"response": "ok"}

        @self.router.get("/server_stats")
        @self.base_endpoint
        async def server_stats(
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ) -> WrappedServerStatsResponse:
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only an authorized user can call the `server_stats` endpoint.",
                    403,
                )
            return {
                "start_time": self.start_time.isoformat(),
                "uptime_seconds": (
                    datetime.now(timezone.utc) - self.start_time
                ).total_seconds(),
                "cpu_usage": psutil.cpu_percent(),
                "memory_usage": psutil.virtual_memory().percent,
            }

        @self.router.post("/update_prompt")
        @self.base_endpoint
        async def update_prompt_app(
            name: str = Body(..., description="Prompt name"),
            template: Optional[str] = Body(
                None, description="Prompt template"
            ),
            input_types: Optional[dict[str, str]] = Body(
                {}, description="Input types"
            ),
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ) -> WrappedPromptResponse:
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can call the `update_prompt` endpoint.",
                    403,
                )

            result = await self.engine.aupdate_prompt(
                name, template, input_types
            )
            return result

        @self.router.get("/analytics")
        @self.base_endpoint
        async def get_analytics_app(
            filter_criteria: Optional[Json[dict]] = Query({}),
            analysis_types: Optional[Json[dict]] = Query({}),
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ) -> WrappedAnalyticsResponse:
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can call the `analytics` endpoint.", 403
                )

            try:
                result = await self.engine.aanalytics(
                    filter_criteria=LogFilterCriteria(filters=filter_criteria),
                    analysis_types=AnalysisTypes(
                        analysis_types=analysis_types
                    ),
                )
                return result
            except json.JSONDecodeError as e:
                raise R2RException(
                    f"Invalid JSON in query parameters: {str(e)}", 400
                )
            except ValueError as e:
                raise R2RException(
                    f"Invalid data in query parameters: {str(e)}", 400
                )

        # TODO: should we add a message to the response model with warnings i.e. if the max_runs passed in violates the max_runs limit?
        @self.router.get("/logs")
        @self.base_endpoint
        async def logs_app(
            run_type_filter: Optional[str] = Query(""),
            max_runs: int = Query(100, ge=1, le=1000),
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ) -> WrappedLogResponse:
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can call the `logs` endpoint.", 403
                )

            return await self.engine.alogs(
                run_type_filter=run_type_filter,
                max_runs=min(max(max_runs, 1), 1000),
            )

        @self.router.get("/app_settings")
        @self.base_endpoint
        async def app_settings(
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ) -> WrappedAppSettingsResponse:
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can call the `app_settings` endpoint.",
                    403,
                )
            return await self.engine.aapp_settings()

        @self.router.post("/score_completion")
        @self.base_endpoint
        async def score_completion(
            message_id: str = Body(..., description="Message ID"),
            score: float = Body(..., description="Completion score"),
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ) -> WrappedScoreCompletionResponse:
            message_uuid = UUID(message_id)
            return await self.engine.ascore_completion(
                message_id=message_uuid, score=score
            )

        @self.router.get("/users_overview")
        @self.base_endpoint
        async def users_overview_app(
            user_ids: Optional[list[str]] = Query([]),
            offset: int = Query(0, ge=0),
            limit: int = Query(100, ge=1, le=1000),
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ) -> WrappedUserOverviewResponse:
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can call the `users_overview` endpoint.",
                    403,
                )

            user_uuids = (
                [UUID(user_id) for user_id in user_ids] if user_ids else None
            )

            return await self.engine.ausers_overview(
                user_ids=user_uuids, offset=offset, limit=limit
            )

        @self.router.delete("/delete", status_code=204)
        @self.base_endpoint
        async def delete_app(
            filters: str = Query(..., description="JSON-encoded filters"),
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ) -> None:
            filters_dict = json.loads(filters) if filters else None
            return await self.engine.adelete(filters=filters_dict)

        @self.router.get("/documents_overview")
        @self.base_endpoint
        async def documents_overview_app(
            document_ids: list[str] = Query([]),
            offset: int = Query(0, ge=0),
            limit: int = Query(100, ge=1, le=1000),
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ) -> WrappedDocumentOverviewResponse:
            request_user_ids = (
                None if auth_user.is_superuser else [auth_user.id]
            )
            document_uuids = [
                UUID(document_id) for document_id in document_ids
            ]
            result = await self.engine.adocuments_overview(
                user_ids=request_user_ids,
                group_ids=auth_user.group_ids,
                document_ids=document_uuids,
                offset=offset,
                limit=limit,
            )
            return result

        @self.router.get("/document_chunks/{document_id}")
        @self.base_endpoint
        async def document_chunks_app(
            document_id: str = Path(...),
            offset: Optional[int] = Query(0, ge=0),
            limit: Optional[int] = Query(100, ge=0),
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ) -> WrappedDocumentChunkResponse:
            document_uuid = UUID(document_id)
            chunks = await self.engine.adocument_chunks(
                document_uuid, offset, limit
            )

            if not chunks:
                raise R2RException(
                    "No chunks found for the given document ID.",
                    404,
                )

            is_owner = str(chunks[0].get("user_id")) == str(auth_user.id)

            if not is_owner and not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can arbitrarily call document_chunks.",
                    403,
                )

            return chunks

        @self.router.get("/inspect_knowledge_graph")
        @self.base_endpoint
        async def inspect_knowledge_graph(
            offset: int = 0,
            limit: int = 100,
            print_descriptions: bool = False,
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ) -> WrappedKnowledgeGraphResponse:
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can call the `inspect_knowledge_graph` endpoint.",
                    403,
                )
            return await self.engine.ainspect_knowledge_graph(
                offset=offset,
                limit=limit,
                print_descriptions=print_descriptions,
            )

        @self.router.get("/groups_overview")
        @self.base_endpoint
        async def groups_overview_app(
            group_ids: Optional[list[str]] = Query(None),
            offset: Optional[int] = Query(0, ge=0),
            limit: Optional[int] = Query(100, ge=1, le=1000),
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ) -> WrappedGroupOverviewResponse:
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can call the `groups_overview` endpoint.",
                    403,
                )

            group_uuids = (
                [UUID(group_id) for group_id in group_ids]
                if group_ids
                else None
            )
            return await self.engine.agroups_overview(
                group_ids=group_uuids, offset=offset, limit=limit
            )

        @self.router.post("/create_group")
        @self.base_endpoint
        async def create_group_app(
            name: str = Body(..., description="Group name"),
            description: Optional[str] = Body(
                "", description="Group description"
            ),
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ) -> WrappedGroupResponse:
            if not auth_user.is_superuser:
                raise R2RException("Only a superuser can create groups.", 403)
            return await self.engine.acreate_group(name, description)

        @self.router.get("/get_group/{group_id}")
        @self.base_endpoint
        async def get_group_app(
            group_id: str = Path(..., description="Group ID"),
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ) -> WrappedGroupResponse:
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can get group details.", 403
                )
            group_uuid = UUID(group_id)
            result = await self.engine.aget_group(group_uuid)
            return result

        @self.router.put("/update_group")
        @self.base_endpoint
        async def update_group_app(
            group_id: str = Body(..., description="Group ID"),
            name: Optional[str] = Body(None, description="Updated group name"),
            description: Optional[str] = Body(
                None, description="Updated group description"
            ),
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ) -> WrappedGroupResponse:
            if not auth_user.is_superuser:
                raise R2RException("Only a superuser can update groups.", 403)
            group_uuid = UUID(group_id)
            return await self.engine.aupdate_group(
                group_uuid, name, description
            )

        @self.router.delete("/delete_group/{group_id}")
        @self.base_endpoint
        async def delete_group_app(
            group_id: str = Path(..., description="Group ID"),
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ):
            if not auth_user.is_superuser:
                raise R2RException("Only a superuser can delete groups.", 403)
            group_uuid = UUID(group_id)
            return await self.engine.adelete_group(group_uuid)

        @self.router.get("/list_groups")
        @self.base_endpoint
        async def list_groups_app(
            offset: int = Query(0, ge=0),
            limit: int = Query(100, ge=1, le=1000),
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ) -> WrappedGroupListResponse:
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can list all groups.", 403
                )
            return await self.engine.alist_groups(
                offset=offset, limit=min(max(limit, 1), 1000)
            )

        @self.router.post("/add_user_to_group")
        @self.base_endpoint
        async def add_user_to_group_app(
            user_id: str = Body(..., description="User ID"),
            group_id: str = Body(..., description="Group ID"),
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ) -> WrappedAddUserResponse:
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can add users to groups.", 403
                )
            user_uuid = UUID(user_id)
            group_uuid = UUID(group_id)
            return await self.engine.aadd_user_to_group(user_uuid, group_uuid)

        @self.router.post("/remove_user_from_group")
        @self.base_endpoint
        async def remove_user_from_group_app(
            user_id: str = Body(..., description="User ID"),
            group_id: str = Body(..., description="Group ID"),
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can remove users from groups.", 403
                )
            user_uuid = UUID(user_id)
            group_uuid = UUID(group_id)
            await self.engine.aremove_user_from_group(user_uuid, group_uuid)
            return None

        # TODO - Proivde response model
        @self.router.get("/get_users_in_group/{group_id}")
        @self.base_endpoint
        async def get_users_in_group_app(
            group_id: str = Path(..., description="Group ID"),
            offset: int = Query(0, ge=0, description="Pagination offset"),
            limit: int = Query(
                100, ge=1, le=1000, description="Pagination limit"
            ),
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can get users in a group.", 403
                )
            group_uuid = UUID(group_id)
            return await self.engine.aget_users_in_group(
                group_id=group_uuid,
                offset=offset,
                limit=min(max(limit, 1), 1000),
            )

        @self.router.get("/user_groups/{user_id}")
        @self.base_endpoint
        async def get_groups_for_user_app(
            user_id: str = Path(..., description="User ID"),
            offset: int = Query(0, ge=0, description="Pagination offset"),
            limit: int = Query(
                100, ge=1, le=1000, description="Pagination limit"
            ),
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can get groups for a user.", 403
                )
            user_uuid = UUID(user_id)
            return await self.engine.aget_groups_for_user(
                user_uuid, offset, limit
            )

        @self.router.post("/assign_document_to_group")
        @self.base_endpoint
        async def assign_document_to_group_app(
            document_id: str = Body(..., description="Document ID"),
            group_id: str = Body(..., description="Group ID"),
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ):
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can assign documents to groups.", 403
                )
            document_uuid = UUID(document_id)
            group_uuid = UUID(group_id)
            return await self.engine.aassign_document_to_group(
                document_uuid, group_uuid
            )

        @self.router.post("/remove_document_from_group")
        @self.base_endpoint
        async def remove_document_from_group_app(
            document_id: str = Body(..., description="Document ID"),
            group_id: str = Body(..., description="Group ID"),
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ) -> None:
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can remove documents from groups.", 403
                )
            document_uuid = UUID(document_id)
            group_uuid = UUID(group_id)
            await self.engine.aremove_document_from_group(
                document_uuid, group_uuid
            )
            return None

        @self.router.get("/document_groups/{document_id}")
        @self.base_endpoint
        async def document_groups_app(
            document_id: str = Path(..., description="Document ID"),
            offset: int = Query(0, ge=0),
            limit: int = Query(100, ge=1, le=1000),
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ) -> WrappedGroupListResponse:
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can get the groups belonging to a document.",
                    403,
                )
            return await self.engine.adocument_groups(
                document_id, offset, limit
            )

        @self.router.get("/group/{group_id}/documents")
        @self.base_endpoint
        async def documents_in_group_app(
            group_id: str = Path(..., description="Group ID"),
            offset: int = Query(0, ge=0),
            limit: int = Query(100, ge=1, le=1000),
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ) -> WrappedDocumentOverviewResponse:
            if not auth_user.is_superuser:
                raise R2RException(
                    "Only a superuser can get documents in a group.", 403
                )
            group_uuid = UUID(group_id)
            return await self.engine.adocuments_in_group(
                group_uuid, offset, limit
            )



import asyncio
import logging
from asyncio import Queue
from typing import Any, Optional

from core.base.api.models.auth.responses import UserResponse
from core.base.logging.run_logger import RunLoggingSingleton
from core.base.logging.run_manager import RunManager, manage_run
from core.base.pipeline.base_pipeline import AsyncPipeline, dequeue_requests
from core.base.pipes.base_pipe import AsyncPipe, AsyncState
from core.base.providers.chunking import ChunkingProvider

logger = logging.getLogger(__name__)


class IngestionPipeline(AsyncPipeline):
    """A pipeline for ingestion."""

    def __init__(
        self,
        pipe_logger: Optional[RunLoggingSingleton] = None,
        run_manager: Optional[RunManager] = None,
    ):
        super().__init__(pipe_logger, run_manager)
        self.parsing_pipe = None
        self.embedding_pipeline = None
        self.kg_pipeline = None

    async def run(
        self,
        input: Any,
        state: Optional[AsyncState] = None,
        stream: bool = False,
        run_manager: Optional[RunManager] = None,
        chunking_settings: Optional[ChunkingProvider] = None,
        user: Optional[UserResponse] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self.state = state or AsyncState()
        if self.parsing_pipe is None:
            raise ValueError(
                "parsing_pipe must be set before running the ingestion pipeline"
            )
        if self.chunking_pipe is None:
            raise ValueError(
                "chunking_pipe must be set before running the ingestion pipeline"
            )
        if self.embedding_pipeline is None and self.kg_pipeline is None:
            raise ValueError(
                "At least one of embedding_pipeline or kg_pipeline must be set before running the ingestion pipeline"
            )

        # Use queues to pass data between pipes and duplicate for each pipeline
        parsing_to_chunking_queue = Queue()
        embedding_queue = Queue()
        kg_queue = Queue()

        async def process_documents():
            async for parsed_doc in await self.parsing_pipe.run(
                self.parsing_pipe.Input(message=input),
                state,
                run_manager,
                *args,
                **kwargs,
            ):
                await parsing_to_chunking_queue.put(parsed_doc)
            await parsing_to_chunking_queue.put(None)

            async for chunked_doc in await self.chunking_pipe.run(
                self.chunking_pipe.Input(
                    message=dequeue_requests(parsing_to_chunking_queue)
                ),
                state,
                run_manager,
                chunking_settings=chunking_settings,
                *args,
                **kwargs,
            ):
                if self.embedding_pipeline:
                    await embedding_queue.put(chunked_doc)
                if self.kg_pipeline:
                    await kg_queue.put(chunked_doc)
            await embedding_queue.put(None)
            await kg_queue.put(None)

        # Start the document processing
        process_task = asyncio.create_task(process_documents())

        # Start the embedding and KG pipelines in parallel
        tasks = []
        if self.embedding_pipeline:
            embedding_task = asyncio.create_task(
                self.embedding_pipeline.run(
                    dequeue_requests(embedding_queue),
                    state,
                    stream,
                    run_manager,
                    *args,
                    **kwargs,
                )
            )
            tasks.append(embedding_task)

        if self.kg_pipeline:
            kg_task = asyncio.create_task(
                self.kg_pipeline.run(
                    dequeue_requests(kg_queue),
                    state,
                    stream,
                    run_manager,
                    *args,
                    **kwargs,
                )
            )
            tasks.append(kg_task)

        # Wait for all tasks to complete
        await process_task
        results = await asyncio.gather(*tasks)

        return {
            "embedding_pipeline_output": (
                results[0] if self.embedding_pipeline else None
            ),
            "kg_pipeline_output": (results[-1] if self.kg_pipeline else None),
        }

    def add_pipe(
        self,
        pipe: AsyncPipe,
        add_upstream_outputs: Optional[list[dict[str, str]]] = None,
        parsing_pipe: bool = False,
        kg_pipe: bool = False,
        chunking_pipe: bool = False,
        embedding_pipe: bool = False,
        *args,
        **kwargs,
    ) -> None:
        logger.debug(
            f"Adding pipe {pipe.config.name} to the IngestionPipeline"
        )

        if parsing_pipe:
            self.parsing_pipe = pipe
        elif chunking_pipe:
            self.chunking_pipe = pipe
        elif kg_pipe:
            if not self.kg_pipeline:
                self.kg_pipeline = AsyncPipeline()
            self.kg_pipeline.add_pipe(
                pipe, add_upstream_outputs, *args, **kwargs
            )
        elif embedding_pipe:
            if not self.embedding_pipeline:
                self.embedding_pipeline = AsyncPipeline()
            self.embedding_pipeline.add_pipe(
                pipe, add_upstream_outputs, *args, **kwargs
            )
        else:
            raise ValueError("Pipe must be a parsing, embedding, or KG pipe")



"""Abstractions for documents and their extractions."""

import json
import logging
from datetime import datetime
from enum import Enum
from typing import Optional, Union
from uuid import NAMESPACE_DNS, UUID, uuid4

from pydantic import BaseModel, Field

logger = logging.getLogger(__name__)

DataType = Union[str, bytes]


class DocumentStatus(str, Enum):
    """Status of document processing."""

    PROCESSING = "processing"
    # TODO - Extend support for `partial-failure`
    # PARTIAL_FAILURE = "partial-failure"
    FAILURE = "failure"
    SUCCESS = "success"


class DocumentType(str, Enum):
    """Types of documents that can be stored."""

    CSV = "csv"
    DOCX = "docx"
    HTML = "html"
    HTM = "htm"
    JSON = "json"
    MD = "md"
    PDF = "pdf"
    PPTX = "pptx"
    TXT = "txt"
    XLSX = "xlsx"
    GIF = "gif"
    PNG = "png"
    JPG = "jpg"
    JPEG = "jpeg"
    SVG = "svg"
    MP3 = "mp3"
    MP4 = "mp4"


class Document(BaseModel):
    id: UUID = Field(default_factory=uuid4)
    group_ids: list[UUID]
    user_id: UUID

    type: DocumentType
    data: Union[str, bytes]
    metadata: dict

    def __init__(self, *args, **kwargs):
        doc_type = kwargs.get("type")
        if isinstance(doc_type, str):
            kwargs["type"] = DocumentType(doc_type)

        # Generate UUID based on the hash of the data
        if "id" not in kwargs:
            data = kwargs["data"]
            if isinstance(data, bytes):
                data_str = data.decode("utf-8", errors="ignore")
            else:
                data_str = data
            data_hash = uuid4(NAMESPACE_DNS, data_str)
            kwargs["id"] = data_hash  # Set the id based on the data hash

        super().__init__(*args, **kwargs)

    class Config:
        arbitrary_types_allowed = True
        json_encoders = {
            UUID: str,
            bytes: lambda v: v.decode("utf-8", errors="ignore"),
        }


class DocumentInfo(BaseModel):
    """Base class for document information handling."""

    id: UUID
    group_ids: list[UUID]
    user_id: UUID
    type: DocumentType
    metadata: dict
    title: Optional[str] = None
    version: str
    size_in_bytes: int
    status: DocumentStatus = DocumentStatus.PROCESSING
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None

    def convert_to_db_entry(self):
        """Prepare the document info for database entry, extracting certain fields from metadata."""
        now = datetime.now()

        return {
            "document_id": self.id,
            "group_ids": self.group_ids,
            "user_id": self.user_id,
            "type": self.type,
            "metadata": json.dumps(self.metadata),
            "title": self.title or "N/A",
            "version": self.version,
            "size_in_bytes": self.size_in_bytes,
            "status": self.status,
            "created_at": self.created_at or now,
            "updated_at": self.updated_at or now,
        }


class DocumentExtraction(BaseModel):
    """An extraction from a document."""

    id: UUID
    document_id: UUID
    group_ids: list[UUID]
    user_id: UUID
    data: DataType
    metadata: dict


class DocumentFragment(BaseModel):
    """A fragment extracted from a document."""

    id: UUID
    extraction_id: UUID
    document_id: UUID
    user_id: UUID
    group_ids: list[UUID]
    data: DataType
    metadata: dict



import logging
from collections import defaultdict
from datetime import datetime
from typing import Any, Optional
from uuid import UUID

from fastapi import UploadFile

from core.base import (
    Document,
    DocumentInfo,
    DocumentType,
    R2RDocumentProcessingError,
    R2RException,
    RunLoggingSingleton,
    RunManager,
    generate_user_document_id,
    increment_version,
    to_async_generator,
)
from core.base.api.models import IngestionResponse
from core.base.providers import ChunkingProvider
from core.telemetry.telemetry_decorator import telemetry_event

from ...base.api.models.auth.responses import UserResponse
from ..abstractions import R2RAgents, R2RPipelines, R2RProviders
from ..assembly.config import R2RConfig
from .base import Service

logger = logging.getLogger(__name__)
MB_CONVERSION_FACTOR = 1024 * 1024
STARTING_VERSION = "v0"


class IngestionService(Service):
    def __init__(
        self,
        config: R2RConfig,
        providers: R2RProviders,
        pipelines: R2RPipelines,
        agents: R2RAgents,
        run_manager: RunManager,
        logging_connection: RunLoggingSingleton,
    ):
        super().__init__(
            config,
            providers,
            pipelines,
            agents,
            run_manager,
            logging_connection,
        )

    @telemetry_event("IngestFiles")
    async def ingest_files(
        self,
        files: list[UploadFile],
        user: UserResponse,
        metadatas: Optional[list[dict]] = None,
        document_ids: Optional[list[UUID]] = None,
        chunking_provider: Optional[ChunkingProvider] = None,
        *args: Any,
        **kwargs: Any,
    ) -> IngestionResponse:
        if not files:
            raise R2RException(
                status_code=400, message="No files provided for ingestion."
            )

        try:
            documents = []
            for iteration, file in enumerate(files):
                if not file.filename:
                    raise R2RException(
                        status_code=400, message="File name not provided."
                    )

                document_metadata = metadatas[iteration] if metadatas else {}

                # document id is dynamically generated from the filename and user id, unless explicitly provided
                document_id = (
                    document_ids[iteration]
                    if document_ids
                    else generate_user_document_id(file.filename, user.id)
                )
                document = self._file_to_document(
                    file, user, document_id, document_metadata
                )
                documents.append(document)
            # ingests all documents in parallel
            return await self.ingest_documents(
                documents,
                chunking_provider=chunking_provider,
                *args,
                **kwargs,
            )

        finally:
            for file in files:
                file.file.close()

    @telemetry_event("UpdateFiles")
    async def update_files(
        self,
        files: list[UploadFile],
        user: UserResponse,
        document_ids: Optional[list[UUID]],
        metadatas: Optional[list[dict]] = None,
        chunking_provider: Optional[ChunkingProvider] = None,
        *args: Any,
        **kwargs: Any,
    ) -> IngestionResponse:
        if not files:
            raise R2RException(
                status_code=400, message="No files provided for update."
            )
        if not self.providers.database:
            raise R2RException(
                status_code=501,
                message="Database provider is not available for updating documents.",
            )
        try:
            if document_ids:
                if len(document_ids) != len(files):
                    raise R2RException(
                        status_code=400,
                        message="Number of ids does not match number of files.",
                    )
            else:
                document_ids = [
                    generate_user_document_id(file.filename, user.id)
                    for file in files
                ]
            # Only superusers can modify arbitrary document ids, which this gate guarantees in conjuction with the check that follows
            documents_overview = (
                (
                    self.providers.database.relational.get_documents_overview(
                        filter_document_ids=document_ids,
                    )
                )
                if user.is_superuser
                else self.providers.database.relational.get_documents_overview(
                    filter_document_ids=document_ids, filter_user_ids=[user.id]
                )
            )

            if len(documents_overview) != len(files):
                raise R2RException(
                    status_code=404,
                    message="One or more documents was not found.",
                )

            documents = []
            new_versions = []

            for it, (file, doc_id, doc_info) in enumerate(
                zip(files, document_ids, documents_overview)
            ):
                if not doc_info:
                    raise R2RException(
                        status_code=404,
                        message=f"Document with id {doc_id} not found.",
                    )

                new_version = increment_version(doc_info.version)
                new_versions.append(new_version)

                updated_metadata = (
                    metadatas[it] if metadatas else doc_info.metadata
                )
                updated_metadata["title"] = (
                    updated_metadata.get("title", None)
                    or file.filename.split("/")[-1]
                )

                document = self._file_to_document(
                    file, user, doc_id, updated_metadata
                )
                documents.append(document)

            ingestion_results = await self.ingest_documents(
                documents,
                chunking_provider=chunking_provider,
                versions=new_versions,
                *args,
                **kwargs,
            )

            for doc_id, old_version in zip(
                document_ids,
                [doc_info.version for doc_info in documents_overview],
            ):
                self.providers.database.vector.delete(
                    filters={
                        "document_id": {"$eq": doc_id},
                        "version": {"$eq": old_version},
                    }
                )
                self.providers.database.relational.delete_from_documents_overview(
                    doc_id, old_version
                )

            return ingestion_results

        finally:
            for file in files:
                file.file.close()

    async def ingest_documents(
        self,
        documents: list[Document],
        versions: Optional[list[str]] = None,
        chunking_provider: Optional[ChunkingProvider] = None,
        *args: Any,
        **kwargs: Any,
    ):
        if len(documents) == 0:
            raise R2RException(
                status_code=400, message="No documents provided for ingestion."
            )

        if chunking_provider is None:
            logger.info("No chunking provider specified. Using default.")
        else:
            logger.info(
                f"Using custom chunking provider: {type(chunking_provider).__name__}"
            )

        document_infos = []
        skipped_documents = []
        processed_documents = {}
        duplicate_documents = defaultdict(list)

        if not all(doc.id for doc in documents):
            raise R2RException(
                status_code=400, message="All documents must have an ID."
            )
        user_id = documents[0].user_id

        existing_documents = (
            (
                self.providers.database.relational.get_documents_overview(
                    filter_user_ids=[user_id]
                )
            )
            if self.providers.database
            else []
        )

        existing_document_info = {
            doc_info.id: doc_info for doc_info in existing_documents
        }

        for iteration, document in enumerate(documents):
            version = versions[iteration] if versions else STARTING_VERSION

            # Check for duplicates within the current batch
            if document.id in processed_documents:
                duplicate_documents[document.id].append(
                    document.metadata.get("title", str(document.id))
                )
                continue

            if (
                document.id in existing_document_info
                # apply `geq` check to prevent re-ingestion of updated documents
                and (existing_document_info[document.id].version >= version)
                and existing_document_info[document.id].status == "success"
            ):
                logger.error(
                    f"Document with ID {document.id} was already successfully processed."
                )
                if len(documents) == 1:
                    raise R2RException(
                        status_code=409,
                        message=f"Document with ID {document.id} was already successfully processed.",
                    )
                skipped_documents.append(
                    {
                        "id": document.id,
                        "title": document.metadata.get("title", "N/A"),
                    }
                )
                continue

            now = datetime.now()
            document_info_metadata = document.metadata.copy()
            title = document_info_metadata.pop("title", "N/A")

            document_infos.append(
                DocumentInfo(
                    id=document.id,
                    user_id=document.user_id,
                    group_ids=document.group_ids,
                    type=document.type,
                    metadata=document_info_metadata,
                    title=title,
                    version=version,
                    size_in_bytes=len(document.data),
                    status="processing",
                    created_at=now,
                    updated_at=now,
                )
            )

            processed_documents[document.id] = document.metadata.get(
                "title", str(document.id)
            )
            # Add version to metadata to propagate through pipeline
            document.metadata["version"] = version

        if duplicate_documents:
            duplicate_details = [
                f"{doc_id}: {', '.join(titles)}"
                for doc_id, titles in duplicate_documents.items()
            ]
            warning_message = f"Duplicate documents detected: {'; '.join(duplicate_details)}. These duplicates were skipped."
            raise R2RException(status_code=418, message=warning_message)

        if skipped_documents and len(skipped_documents) == len(documents):
            logger.error("All provided documents already exist.")
            raise R2RException(
                status_code=409,
                message="All provided documents already exist. Use the `update_files` endpoint instead to update these documents.",
            )

        # Insert pending document info
        if self.providers.database:
            self.providers.database.relational.upsert_documents_overview(
                document_infos
            )

        ingestion_results = await self.pipelines.ingestion_pipeline.run(
            input=to_async_generator(
                [
                    doc
                    for doc in documents
                    if doc.id
                    not in [skipped["id"] for skipped in skipped_documents]
                ],
            ),
            run_manager=self.run_manager,
            *args,
            **kwargs,
        )

        return await self._process_ingestion_results(
            ingestion_results,
            document_infos,
            skipped_documents,
        )

    def _file_to_document(
        self,
        file: UploadFile,
        user: UserResponse,
        document_id: UUID,
        metadata: dict,
    ) -> Document:
        file_extension = file.filename.split(".")[-1].lower()
        if file_extension.upper() not in DocumentType.__members__:
            raise R2RException(
                status_code=415,
                message=f"'{file_extension}' is not a valid DocumentType.",
            )

        document_title = metadata.get("title") or file.filename.split("/")[-1]
        metadata["title"] = document_title

        return Document(
            id=document_id,
            group_ids=metadata.get("group_ids", []),
            user_id=user.id,
            type=DocumentType[file_extension.upper()],
            data=file.file.read(),
            metadata=metadata,
        )

    async def _process_ingestion_results(
        self,
        ingestion_results: dict,
        document_infos: list[DocumentInfo],
        skipped_documents: list[dict[str, str]],
    ):
        skipped_ids = [ele["id"] for ele in skipped_documents]
        failed_ids = []
        successful_ids = []

        results = {}
        if ingestion_results["embedding_pipeline_output"]:
            results = dict(ingestion_results["embedding_pipeline_output"])
            for doc_id, error in results.items():
                if isinstance(error, R2RDocumentProcessingError):
                    logger.error(
                        f"Error processing document with ID {error.document_id}: {error.message}"
                    )
                    failed_ids.append(error.document_id)
                elif isinstance(error, Exception):
                    logger.error(f"Error processing document: {error}")
                    failed_ids.append(doc_id)
                else:
                    successful_ids.append(doc_id)

        documents_to_upsert = []
        for document_info in document_infos:
            if document_info.id not in skipped_ids:
                if document_info.id in failed_ids:
                    document_info.status = "failure"
                elif document_info.id in successful_ids:
                    document_info.status = "success"
                documents_to_upsert.append(document_info)

        if documents_to_upsert:
            self.providers.database.relational.upsert_documents_overview(
                documents_to_upsert
            )

        # TODO - modify ingestion service so that at end we write out number
        # of vectors produced or the error message to document info
        # THEN, return updated document infos here
        return {
            "processed_documents": [
                document
                for document in document_infos
                if document.id in successful_ids
            ],
            "failed_documents": [
                {
                    "document_id": document_id,
                    "result": str(results[document_id]),
                }
                for document_id in failed_ids
            ],
            "skipped_documents": skipped_ids,
        }



# Copyright (c) 2024 Microsoft Corporation.
# Licensed under the MIT License

"""A module for clustering entities and triples into communities using hierarchical Leiden algorithm."""

import asyncio
import json
import logging
from typing import Any, AsyncGenerator, Optional
from uuid import UUID

import networkx as nx

from core.base import (
    AsyncPipe,
    AsyncState,
    Community,
    CompletionProvider,
    EmbeddingProvider,
    KGEnrichmentSettings,
    KGProvider,
    PipeType,
    PromptProvider,
    RunLoggingSingleton,
    Triple,
)

logger = logging.getLogger(__name__)


class KGClusteringPipe(AsyncPipe):
    """
    Clusters entities and triples into communities within the knowledge graph using hierarchical Leiden algorithm.
    """

    def __init__(
        self,
        kg_provider: KGProvider,
        llm_provider: CompletionProvider,
        prompt_provider: PromptProvider,
        embedding_provider: EmbeddingProvider,
        pipe_logger: Optional[RunLoggingSingleton] = None,
        type: PipeType = PipeType.OTHER,
        config: Optional[AsyncPipe.PipeConfig] = None,
        *args,
        **kwargs,
    ):
        """
        Initializes the KG clustering pipe with necessary components and configurations.
        """
        super().__init__(
            pipe_logger=pipe_logger,
            type=type,
            config=config or AsyncPipe.PipeConfig(name="kg_cluster_pipe"),
        )
        self.kg_provider = kg_provider
        self.llm_provider = llm_provider
        self.prompt_provider = prompt_provider
        self.embedding_provider = embedding_provider

    def _compute_leiden_communities(
        self,
        graph: nx.Graph,
        settings: KGEnrichmentSettings,
    ) -> dict[int, dict[str, int]]:
        """Compute Leiden communities."""
        try:
            from graspologic.partition import hierarchical_leiden

            community_mapping = hierarchical_leiden(
                graph, **settings.leiden_params
            )
            results: dict[int, dict[str, int]] = {}
            for partition in community_mapping:
                results[partition.level] = results.get(partition.level, {})
                results[partition.level][partition.node] = partition.cluster

            return results
        except ImportError as e:
            raise ImportError("Please install the graspologic package.") from e

    async def cluster_kg(
        self,
        triples: list[Triple],
        settings: KGEnrichmentSettings = KGEnrichmentSettings(),
    ) -> list[Community]:
        """
        Clusters the knowledge graph triples into communities using hierarchical Leiden algorithm.
        """

        G = nx.Graph()
        for triple in triples:
            G.add_edge(
                triple.subject,
                triple.object,
                weight=triple.weight,
                description=triple.description,
                predicate=triple.predicate,
                id=f"{triple.subject}->{triple.predicate}->{triple.object}",
            )

        hierarchical_communities = self._compute_leiden_communities(
            G, settings=settings
        )

        community_details = {}

        for level, level_communities in hierarchical_communities.items():
            for node, cluster in level_communities.items():
                if f"{level}_{cluster}" not in community_details:
                    community_details[f"{level}_{cluster}"] = Community(
                        id=f"{level}_{cluster}",
                        level=str(level),
                        entity_ids=[],
                        relationship_ids=[],
                        short_id=f"{level}.{cluster}",
                        title=f"Community {level}.{cluster}",
                        attributes={
                            "name": f"Community {level}.{cluster}",
                            "community_report": None,
                        },
                    )

                community_details[f"{level}_{cluster}"].entity_ids.append(node)
                for neighbor in G.neighbors(node):
                    edge_info = G.get_edge_data(node, neighbor)
                    if edge_info and edge_info.get("id"):
                        community_details[
                            f"{level}_{cluster}"
                        ].relationship_ids.append(edge_info.get("id"))

        async def async_iterate_dict(dictionary):
            for key, value in dictionary.items():
                yield key, value

        async def process_community(community_key, community):
            input_text = """

                Entities:
                {entities}

                Relationships:
                {relationships}

            """

            entities_info = self.kg_provider.get_entities(community.entity_ids)
            entities_info = "\n".join(
                [
                    f"{entity.name}, {entity.description}"
                    for entity in entities_info
                ]
            )

            relationships_info = self.kg_provider.get_triples(
                community.relationship_ids
            )
            relationships_info = "\n".join(
                [
                    f"{relationship.subject}, {relationship.object}, {relationship.predicate}, {relationship.description}"
                    for relationship in relationships_info
                ]
            )

            input_text = input_text.format(
                entities=entities_info, relationships=relationships_info
            )

            description = await self.llm_provider.aget_completion(
                messages=self.prompt_provider._get_message_payload(
                    task_prompt_name="graphrag_community_reports",
                    task_inputs={
                        "input_text": input_text,
                    },
                ),
                generation_config=settings.generation_config,
            )

            description = description.choices[0].message.content
            community.summary = description
            summary_embedding = (
                await self.embedding_provider.async_get_embedding(
                    community.summary
                )
            )
            community.summary_embedding = summary_embedding
            self.kg_provider.upsert_communities([community])
            try:
                summary = json.loads(community.summary)
            except:
                summary = {"title": "_"}
            return {"id": community.id, "title": summary["title"]}

        tasks = []
        async for community_key, community in async_iterate_dict(
            community_details
        ):
            tasks.append(
                asyncio.create_task(
                    process_community(community_key, community)
                )
            )

        total_tasks = len(tasks)
        for i, completed_task in enumerate(asyncio.as_completed(tasks), 1):
            result = await completed_task
            logger.info(
                f"Progress: {i}/{total_tasks} communities completed ({i / total_tasks * 100:.2f}%)"
            )
            yield result

    async def _run_logic(
        self,
        input: AsyncPipe.Input,
        state: AsyncState,
        run_id: UUID,
        *args: Any,
        **kwargs: Any,
    ) -> AsyncGenerator[Community, None]:
        """
        Executes the KG clustering pipe: clustering entities and triples into communities.
        """

        base_dimension = self.embedding_provider.config.base_dimension
        vector_index_fn = self.kg_provider.create_vector_index
        vector_index_fn("__ENTITY__", "name_embedding", base_dimension)
        vector_index_fn("__ENTITY__", "description_embedding", base_dimension)
        vector_index_fn("__RELATIONSHIP__", "description", base_dimension)
        vector_index_fn("__Community__", "summary_embedding", base_dimension)

        all_nodes = []
        async for node in input.message:
            all_nodes.append(node)

        triples = self.kg_provider.get_triples()

        async for community in self.cluster_kg(
            triples, self.kg_provider.config.kg_enrichment_settings
        ):
            yield community



from typing import Optional
from uuid import UUID

from core.base import DocumentInfo, DocumentStatus, DocumentType, R2RException
from core.base.api.models.management.responses import GroupResponse

from .base import DatabaseMixin


class DocumentMixin(DatabaseMixin):

    def create_table(self):
        query = f"""
        CREATE TABLE IF NOT EXISTS {self._get_table_name('document_info')} (
            document_id UUID PRIMARY KEY,
            group_ids UUID[],
            user_id UUID,
            type TEXT,
            metadata JSONB,
            title TEXT,
            version TEXT,
            size_in_bytes INT,
            status TEXT DEFAULT 'processing',
            created_at TIMESTAMPTZ DEFAULT NOW(),
            updated_at TIMESTAMPTZ DEFAULT NOW()
        );
        CREATE INDEX IF NOT EXISTS idx_group_ids_{self.collection_name}
        ON {self._get_table_name('document_info')} USING GIN (group_ids);
        """
        self.execute_query(query)

    def upsert_documents_overview(
        self, documents_overview: list[DocumentInfo]
    ) -> None:
        for document_info in documents_overview:
            query = f"""
            INSERT INTO {self._get_table_name('document_info')}
            (document_id, group_ids, user_id, type, metadata, title, version, size_in_bytes, status, created_at, updated_at)
            VALUES (:document_id, :group_ids, :user_id, :type, :metadata, :title, :version, :size_in_bytes, :status, :created_at, :updated_at)
            ON CONFLICT (document_id) DO UPDATE SET
                group_ids = EXCLUDED.group_ids,
                user_id = EXCLUDED.user_id,
                type = EXCLUDED.type,
                metadata = EXCLUDED.metadata,
                title = EXCLUDED.title,
                version = EXCLUDED.version,
                size_in_bytes = EXCLUDED.size_in_bytes,
                status = EXCLUDED.status,
                updated_at = EXCLUDED.updated_at;
            """
            self.execute_query(query, document_info.convert_to_db_entry())

    def delete_from_documents_overview(
        self, document_id: str, version: Optional[str] = None
    ) -> None:
        query = f"""
            DELETE FROM {self._get_table_name('document_info')}
            WHERE document_id = :document_id
        """
        params = {"document_id": document_id}

        if version is not None:
            query += " AND version = :version"
            params["version"] = version

        self.execute_query(query, params)

    def get_documents_overview(
        self,
        filter_user_ids: Optional[list[UUID]] = None,
        filter_document_ids: Optional[list[UUID]] = None,
        filter_group_ids: Optional[list[UUID]] = None,
        offset: int = 0,
        limit: int = 100,
    ):
        conditions = []
        params = {"offset": offset, "limit": limit}

        if filter_document_ids:
            conditions.append("document_id = ANY(:document_ids)")
            params["document_ids"] = filter_document_ids

        if filter_user_ids:
            conditions.append("user_id = ANY(:user_ids)")
            params["user_ids"] = filter_user_ids

        if filter_group_ids:
            conditions.append("group_ids && :group_ids")
            params["group_ids"] = filter_group_ids

        query = f"""
            SELECT document_id, group_ids, user_id, type, metadata, title, version, size_in_bytes, status, created_at, updated_at
            FROM {self._get_table_name('document_info')}
        """
        if conditions:
            query += " WHERE " + " AND ".join(conditions)

        query += """
            ORDER BY created_at DESC
            OFFSET :offset
            LIMIT :limit
        """

        results = self.execute_query(query, params).fetchall()
        documents = [
            DocumentInfo(
                id=row[0],
                group_ids=row[1],
                user_id=row[2],
                type=DocumentType(row[3]),
                metadata=row[4],
                title=row[5],
                version=row[6],
                size_in_bytes=row[7],
                status=DocumentStatus(row[8]),
                created_at=row[9],
                updated_at=row[10],
            )
            for row in results
        ]

        # Get total count for pagination metadata
        count_query = f"""
            SELECT COUNT(*)
            FROM {self._get_table_name('document_info')}
        """
        if conditions:
            count_query += " WHERE " + " AND ".join(conditions)

        return documents



from typing import Any, Optional

from pydantic import BaseModel, Field

from core.base.abstractions import KGSearchResult, Message, VectorSearchResult
from core.base.api.models.base import ResultsWrapper


class SearchResponse(BaseModel):
    vector_search_results: list[VectorSearchResult] = Field(
        ...,
        description="List of vector search results",
    )
    kg_search_results: Optional[list[KGSearchResult]] = Field(
        None,
        description="Knowledge graph search results, if applicable",
    )

    class Config:
        json_schema_extra = {
            "example": {
                "vector_search_results": [
                    {
                        "fragment_id": "c68dc72e-fc23-5452-8f49-d7bd46088a96",
                        "extraction_id": "3f3d47f3-8baf-58eb-8bc2-0171fb1c6e09",
                        "document_id": "3e157b3a-8469-51db-90d9-52e7d896b49b",
                        "user_id": "2acb499e-8428-543b-bd85-0d9098718220",
                        "group_ids": [],
                        "score": 0.23943702876567796,
                        "text": "Example text from the document",
                        "metadata": {
                            "title": "example_document.pdf",
                            "associatedQuery": "What is the capital of France?",
                        },
                    }
                ],
                "kg_search_results": None,
            }
        }


class RAGResponse(BaseModel):
    completion: Any = Field(
        ...,
        description="The generated completion from the RAG process",
    )
    search_results: SearchResponse = Field(
        ...,
        description="The search results used for the RAG process",
    )

    class Config:
        json_schema_extra = {
            "example": {
                "completion": {
                    "id": "chatcmpl-example123",
                    "choices": [
                        {
                            "finish_reason": "stop",
                            "index": 0,
                            "logprobs": None,
                            "message": {
                                "content": "Paris is the capital of France.",
                                "role": "assistant",
                            },
                        }
                    ],
                },
                "search_results": {
                    "vector_search_results": [
                        {
                            "fragment_id": "c68dc72e-fc23-5452-8f49-d7bd46088a96",
                            "extraction_id": "3f3d47f3-8baf-58eb-8bc2-0171fb1c6e09",
                            "document_id": "3e157b3a-8469-51db-90d9-52e7d896b49b",
                            "user_id": "2acb499e-8428-543b-bd85-0d9098718220",
                            "group_ids": [],
                            "score": 0.23943702876567796,
                            "text": "Paris is the capital and most populous city of France.",
                            "metadata": {
                                "text": "Paris is the capital and most populous city of France.",
                                "title": "france_info.pdf",
                                "associatedQuery": "What is the capital of France?",
                            },
                        }
                    ],
                    "kg_search_results": None,
                },
            }
        }


RAGAgentResponse = list[Message]
# Create wrapped versions of the responses
WrappedSearchResponse = ResultsWrapper[SearchResponse]
WrappedRAGResponse = ResultsWrapper[RAGResponse]
WrappedRAGAgentResponse = ResultsWrapper[RAGAgentResponse]



from .auth.responses import GenericMessageResponse, TokenResponse, UserResponse
from .ingestion.responses import (
    FailedDocument,
    IngestionResponse,
    ProcessedDocument,
    WrappedIngestionResponse,
)
from .management.responses import (
    AnalyticsResponse,
    AppSettingsResponse,
    DocumentChunkResponse,
    DocumentOverviewResponse,
    GroupOverviewResponse,
    GroupResponse,
    KnowledgeGraphResponse,
    LogResponse,
    PromptResponse,
    ScoreCompletionResponse,
    ServerStats,
    UserOverviewResponse,
    WrappedAddUserResponse,
    WrappedAnalyticsResponse,
    WrappedAppSettingsResponse,
    WrappedDocumentChunkResponse,
    WrappedDocumentOverviewResponse,
    WrappedGroupListResponse,
    WrappedGroupOverviewResponse,
    WrappedGroupResponse,
    WrappedKnowledgeGraphResponse,
    WrappedLogResponse,
    WrappedPromptResponse,
    WrappedScoreCompletionResponse,
    WrappedServerStatsResponse,
    WrappedUserOverviewResponse,
)
from .restructure.responses import (
    KGEnrichmentResponse,
    WrappedKGEnrichmentResponse,
)
from .retrieval.responses import (
    RAGAgentResponse,
    RAGResponse,
    SearchResponse,
    WrappedRAGAgentResponse,
    WrappedRAGResponse,
    WrappedSearchResponse,
)

__all__ = [
    # Auth Responses
    "GenericMessageResponse",
    "TokenResponse",
    "UserResponse",
    # Ingestion Responses
    "ProcessedDocument",
    "FailedDocument",
    "IngestionResponse",
    "WrappedIngestionResponse",
    # Restructure Responses
    "KGEnrichmentResponse",
    "WrappedKGEnrichmentResponse",
    # Management Responses
    "PromptResponse",
    "ServerStats",
    "LogResponse",
    "AnalyticsResponse",
    "AppSettingsResponse",
    "ScoreCompletionResponse",
    "UserOverviewResponse",
    "DocumentOverviewResponse",
    "DocumentChunkResponse",
    "KnowledgeGraphResponse",
    "GroupResponse",
    "GroupOverviewResponse",
    "WrappedPromptResponse",
    "WrappedServerStatsResponse",
    "WrappedLogResponse",
    "WrappedAnalyticsResponse",
    "WrappedAppSettingsResponse",
    "WrappedScoreCompletionResponse",
    "WrappedUserOverviewResponse",
    "WrappedDocumentOverviewResponse",
    "WrappedDocumentChunkResponse",
    "WrappedKnowledgeGraphResponse",
    "WrappedGroupResponse",
    "WrappedGroupListResponse",
    "WrappedAddUserResponse",
    "WrappedGroupOverviewResponse",
    # Retrieval Responses
    "SearchResponse",
    "RAGResponse",
    "RAGAgentResponse",
    "WrappedSearchResponse",
    "WrappedRAGResponse",
    "WrappedRAGAgentResponse",
]



"""
This module contains the `DocumentParsingPipe` class, which is responsible for parsing incoming documents into plaintext.
"""

import logging
from typing import AsyncGenerator, Optional, Union
from uuid import UUID

from core.base import (
    AsyncState,
    Document,
    DocumentExtraction,
    ParsingProvider,
    PipeType,
    RunLoggingSingleton,
    generate_id_from_label,
)
from core.base.abstractions.exception import R2RDocumentProcessingError
from core.base.pipes.base_pipe import AsyncPipe

logger = logging.getLogger(__name__)


class ParsingPipe(AsyncPipe):
    class Input(AsyncPipe.Input):
        message: AsyncGenerator[Document, None]

    def __init__(
        self,
        parsing_provider: ParsingProvider,
        pipe_logger: Optional[RunLoggingSingleton] = None,
        type: PipeType = PipeType.INGESTOR,
        config: Optional[AsyncPipe.PipeConfig] = None,
        *args,
        **kwargs,
    ):
        super().__init__(
            pipe_logger=pipe_logger,
            type=type,
            config=config
            or AsyncPipe.PipeConfig(name="default_document_parsing_pipe"),
            *args,
            **kwargs,
        )
        self.parsing_provider = parsing_provider

    async def _parse(
        self,
        document: Document,
        run_id: UUID,
        version: str,
    ) -> AsyncGenerator[
        Union[R2RDocumentProcessingError, DocumentExtraction], None
    ]:
        try:
            async for extraction in self.parsing_provider.parse(document):
                extraction_id = generate_id_from_label(
                    f"{extraction.id}-{version}"
                )
                extraction.id = extraction_id
                # Add version to metadata
                extraction.metadata["version"] = version
                yield extraction
        except Exception as e:
            yield R2RDocumentProcessingError(
                document_id=document.id,
                error_message=f"Error parsing document: {str(e)}",
            )

    async def _run_logic(
        self,
        input: Input,
        state: AsyncState,
        run_id: UUID,
        *args,
        **kwargs,
    ) -> AsyncGenerator[DocumentExtraction, None]:
        async for document in input.message:
            async for result in self._parse(
                document, run_id, document.metadata.get("version", "1.0")
            ):
                yield result



from pydantic import BaseModel, Field

from .llm import GenerationConfig


class KGEnrichmentSettings(BaseModel):
    """Settings for knowledge graph enrichment."""

    max_knowledge_triples: int = Field(
        default=100,
        description="The maximum number of knowledge triples to extract from each chunk.",
    )

    generation_config: GenerationConfig = Field(
        default_factory=GenerationConfig,
        description="Configuration for text generation during graph enrichment.",
    )
    leiden_params: dict = Field(
        default_factory=dict,
        description="Parameters for the Leiden algorithm.",
    )



"""
Defines the 'Collection' class

Importing from the `vecs.collection` directly is not supported.
All public classes, enums, and functions are re-exported by the top level `vecs` module.
"""

from __future__ import annotations

import math
import warnings
from dataclasses import dataclass
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)
from uuid import UUID, uuid4

import psycopg2
import sqlalchemy as sa
from flupy import flu
from nltk.corpus import wordnet
from nltk.stem import SnowballStemmer
from sqlalchemy import (
    Column,
    Index,
    MetaData,
    String,
    Table,
    alias,
    and_,
    cast,
    delete,
    distinct,
    func,
    or_,
    select,
    text,
)
from sqlalchemy.dialects import postgresql
from sqlalchemy.types import Float, UserDefinedType

from core.base import VectorSearchResult
from core.base.abstractions import VectorSearchSettings

from .adapter import Adapter, AdapterContext, NoOp, Record
from .exc import (
    ArgError,
    CollectionAlreadyExists,
    CollectionNotFound,
    FilterError,
    MismatchedDimension,
    Unreachable,
)

if TYPE_CHECKING:
    from vecs.client import Client


class IndexMethod(str, Enum):
    """
    An enum representing the index methods available.

    This class currently only supports the 'ivfflat' method but may
    expand in the future.

    Attributes:
        auto (str): Automatically choose the best available index method.
        ivfflat (str): The ivfflat index method.
        hnsw (str): The hnsw index method.
    """

    auto = "auto"
    ivfflat = "ivfflat"
    hnsw = "hnsw"


class IndexMeasure(str, Enum):
    """
    An enum representing the types of distance measures available for indexing.

    Attributes:
        cosine_distance (str): The cosine distance measure for indexing.
        l2_distance (str): The Euclidean (L2) distance measure for indexing.
        max_inner_product (str): The maximum inner product measure for indexing.
    """

    cosine_distance = "cosine_distance"
    l2_distance = "l2_distance"
    max_inner_product = "max_inner_product"


@dataclass
class IndexArgsIVFFlat:
    """
    A class for arguments that can optionally be supplied to the index creation
    method when building an IVFFlat type index.

    Attributes:
        nlist (int): The number of IVF centroids that the index should use
    """

    n_lists: int


@dataclass
class IndexArgsHNSW:
    """
    A class for arguments that can optionally be supplied to the index creation
    method when building an HNSW type index.

    Ref: https://github.com/pgvector/pgvector#index-options

    Both attributes are Optional in case the user only wants to specify one and
    leave the other as default

    Attributes:
        m (int): Maximum number of connections per node per layer (default: 16)
        ef_construction (int): Size of the dynamic candidate list for
            constructing the graph (default: 64)
    """

    m: Optional[int] = 16
    ef_construction: Optional[int] = 64


INDEX_MEASURE_TO_OPS = {
    # Maps the IndexMeasure enum options to the SQL ops string required by
    # the pgvector `create index` statement
    IndexMeasure.cosine_distance: "vector_cosine_ops",
    IndexMeasure.l2_distance: "vector_l2_ops",
    IndexMeasure.max_inner_product: "vector_ip_ops",
}

INDEX_MEASURE_TO_SQLA_ACC = {
    IndexMeasure.cosine_distance: lambda x: x.cosine_distance,
    IndexMeasure.l2_distance: lambda x: x.l2_distance,
    IndexMeasure.max_inner_product: lambda x: x.max_inner_product,
}


class Vector(UserDefinedType):
    cache_ok = True

    def __init__(self, dim=None):
        super(UserDefinedType, self).__init__()
        self.dim = dim

    def get_col_spec(self, **kw):
        return "VECTOR" if self.dim is None else f"VECTOR({self.dim})"

    def bind_processor(self, dialect):
        def process(value):
            if value is None:
                return value
            if not isinstance(value, list):
                raise ValueError("Expected a list")
            if self.dim is not None and len(value) != self.dim:
                raise ValueError(
                    f"Expected {self.dim} dimensions, not {len(value)}"
                )
            return "[" + ",".join(str(float(v)) for v in value) + "]"

        return process

    def result_processor(self, dialect, coltype):
        return lambda value: (
            value
            if value is None
            else [float(v) for v in value[1:-1].split(",")]
        )

    class comparator_factory(UserDefinedType.Comparator):
        def l2_distance(self, other):
            return self.op("<->", return_type=Float)(other)

        def max_inner_product(self, other):
            return self.op("<#>", return_type=Float)(other)

        def cosine_distance(self, other):
            return self.op("<=>", return_type=Float)(other)


class Collection:
    """
    The `vecs.Collection` class represents a collection of vectors within a PostgreSQL database with pgvector support.
    It provides methods to manage (create, delete, fetch, upsert), index, and perform similarity searches on these vector collections.

    The collections are stored in separate tables in the database, with each vector associated with an identifier and optional metadata.

    Example usage:

        with vecs.create_client(DB_CONNECTION) as vx:
            collection = vx.create_collection(name="docs", dimension=3)
            collection.upsert([("id1", [1, 1, 1], {"key": "value"})])
            # Further operations on 'collection'

    Public Attributes:
        name: The name of the vector collection.
        dimension: The dimension of vectors in the collection.

    Note: Some methods of this class can raise exceptions from the `vecs.exc` module if errors occur.
    """

    COLUMN_VARS = [
        "fragment_id",
        "extraction_id",
        "document_id",
        "user_id",
        "group_ids",
    ]

    def __init__(
        self,
        name: str,
        dimension: int,
        client: Client,
        adapter: Optional[Adapter] = None,
    ):
        """
        Initializes a new instance of the `Collection` class.

        During expected use, developers initialize instances of `Collection` using the
        `vecs.Client` with `vecs.Client.create_collection(...)` rather than directly.

        Args:
            name (str): The name of the collection.
            dimension (int): The dimension of the vectors in the collection.
            client (Client): The client to use for interacting with the database.
        """
        from core.providers.database.vecs.adapter import Adapter

        self.client = client
        self.name = name
        self.dimension = dimension
        self.table = _build_table(name, client.meta, dimension)
        self._index: Optional[str] = None
        self.adapter = adapter or Adapter(steps=[NoOp(dimension=dimension)])

        reported_dimensions = set(
            [
                x
                for x in [
                    dimension,
                    adapter.exported_dimension if adapter else None,
                ]
                if x is not None
            ]
        )
        if len(reported_dimensions) == 0:
            raise ArgError(
                "One of dimension or adapter must provide a dimension"
            )
        elif len(reported_dimensions) > 1:
            raise MismatchedDimension(
                "Mismatch in the reported dimensions of the selected vector collection and embedding model. Correct the selected embedding model or specify a new vector collection by modifying the `POSTGRES_VECS_COLLECTION` environment variable."
            )

    def __repr__(self):
        """
        Returns a string representation of the `Collection` instance.

        Returns:
            str: A string representation of the `Collection` instance.
        """
        return (
            f'vecs.Collection(name="{self.name}", dimension={self.dimension})'
        )

    def __len__(self) -> int:
        """
        Returns the number of vectors in the collection.

        Returns:
            int: The number of vectors in the collection.
        """
        with self.client.Session() as sess:
            with sess.begin():
                stmt = select(func.count()).select_from(self.table)
                return sess.execute(stmt).scalar() or 0

    def _create_if_not_exists(self):
        """
        PRIVATE

        Creates a new collection in the database if it doesn't already exist

        Returns:
            Collection: The found or created collection.
        """
        query = text(
            f"""
        select
            relname as table_name,
            atttypmod as embedding_dim
        from
            pg_class pc
            join pg_attribute pa
                on pc.oid = pa.attrelid
        where
            pc.relnamespace = 'vecs'::regnamespace
            and pc.relkind = 'r'
            and pa.attname = 'vec'
            and not pc.relname ^@ '_'
            and pc.relname = :name
        """
        ).bindparams(name=self.name)
        with self.client.Session() as sess:
            query_result = sess.execute(query).fetchone()

            if query_result:
                _, collection_dimension = query_result
            else:
                collection_dimension = None

        reported_dimensions = set(
            [
                x
                for x in [self.dimension, collection_dimension]
                if x is not None
            ]
        )
        if len(reported_dimensions) > 1:
            raise MismatchedDimension(
                "Mismatch in the reported dimensions of the selected vector collection and embedding model. Correct the selected embedding model or specify a new vector collection by modifying the `POSTGRES_VECS_COLLECTION` environment variable."
            )

        if not collection_dimension:
            self.table.create(self.client.engine)

        return self

    def _create(self):
        """
        PRIVATE

        Creates a new collection in the database. Raises a `vecs.exc.CollectionAlreadyExists`
        exception if a collection with the specified name already exists.

        Returns:
            Collection: The newly created collection.
        """

        collection_exists = self.__class__._does_collection_exist(
            self.client, self.name
        )
        if collection_exists:
            raise CollectionAlreadyExists(
                "Collection with requested name already exists"
            )
        self.table.create(self.client.engine)

        unique_string = str(uuid4()).replace("-", "_")[0:7]
        with self.client.Session() as sess:
            sess.execute(
                text(
                    f"""
                    create index ix_meta_{unique_string}
                      on vecs."{self.table.name}"
                      using gin ( metadata jsonb_path_ops )
                    """
                )
            )

            # Create trigger to update fts column
            sess.execute(
                text(
                    f"""
                CREATE TRIGGER tsvector_update_{unique_string} BEFORE INSERT OR UPDATE
                ON vecs."{self.table.name}" FOR EACH ROW EXECUTE FUNCTION
                tsvector_update_trigger(fts, 'pg_catalog.english', text);
            """
                )
            )
        return self

    def _drop(self):
        """
        PRIVATE

        Deletes the collection from the database. Raises a `vecs.exc.CollectionNotFound`
        exception if no collection with the specified name exists.

        Returns:
            Collection: The deleted collection.
        """
        with self.client.Session() as sess:
            sess.execute(text(f"DROP TABLE IF EXISTS {self.name} CASCADE"))
            sess.commit()

        return self

    def upsert(
        self,
        records: Iterable[Record],
    ) -> None:
        chunk_size = 512

        pipeline = flu(self.adapter(records, AdapterContext("upsert"))).chunk(
            chunk_size
        )

        with self.client.Session() as sess:
            with sess.begin():
                for chunk in pipeline:
                    stmt = postgresql.insert(self.table).values(
                        [
                            {
                                "fragment_id": record[0],
                                "extraction_id": record[1],
                                "document_id": record[2],
                                "user_id": record[3],
                                "group_ids": record[4],
                                "vec": record[5],
                                "text": record[6],
                                "metadata": record[7],
                                "fts": func.to_tsvector(record[6]),
                            }
                            for record in chunk
                        ]
                    )
                    stmt = stmt.on_conflict_do_update(
                        index_elements=[self.table.c.fragment_id],
                        set_=dict(
                            extraction_id=stmt.excluded.extraction_id,
                            document_id=stmt.excluded.document_id,
                            user_id=stmt.excluded.user_id,
                            group_ids=stmt.excluded.group_ids,
                            vec=stmt.excluded.vec,
                            text=stmt.excluded.text,
                            metadata=stmt.excluded.metadata,
                            fts=stmt.excluded.fts,
                        ),
                    )
                    sess.execute(stmt)
        return None

    def fetch(self, fragment_ids: Iterable[UUID]) -> List[Record]:
        """
        Fetches vectors from the collection by their fragment identifiers.

        Args:
            fragment_ids (Iterable[UUID]): An iterable of vector fragment identifiers.

        Returns:
            List[Record]: A list of the fetched vectors.

        Raises:
            ArgError: If fragment_ids is not an iterable of UUIDs.
        """
        if isinstance(fragment_ids, (str, UUID)):
            raise ArgError("fragment_ids must be an iterable of UUIDs")

        chunk_size = 12
        records = []
        with self.client.Session() as sess:
            with sess.begin():
                for id_chunk in flu(fragment_ids).chunk(chunk_size):
                    stmt = select(self.table).where(
                        self.table.c.fragment_id.in_(id_chunk)
                    )
                    chunk_records = sess.execute(stmt)
                    records.extend(chunk_records)
        return records

    def delete(
        self,
        fragment_ids: Optional[Iterable[UUID]] = None,
        filters: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Dict[str, str]]:
        """
        Deletes vectors from the collection by matching filters or fragment_ids.

        Args:
            fragment_ids (Optional[Iterable[UUID]], optional): An iterable of vector fragment identifiers.
            filters (Optional[Dict], optional): Filters to apply to the search. Defaults to None.

        Returns:
            Dict[str, Dict[str, str]]: A dictionary of deleted records, where the key is the fragment_id
            and the value is a dictionary containing 'document_id', 'extraction_id', 'fragment_id', and 'text'.

        Raises:
            ArgError: If neither fragment_ids nor filters are provided, or if both are provided.
        """
        if fragment_ids is None and filters is None:
            raise ArgError("Either fragment_ids or filters must be provided.")

        if fragment_ids is not None and filters is not None:
            raise ArgError(
                "Either fragment_ids or filters must be provided, not both."
            )

        if isinstance(fragment_ids, (str, UUID)):
            raise ArgError("fragment_ids must be an iterable of UUIDs")

        deleted_records = {}

        with self.client.Session() as sess:
            with sess.begin():
                if fragment_ids:
                    for id_chunk in flu(fragment_ids).chunk(12):
                        delete_stmt = (
                            delete(self.table)
                            .where(self.table.c.fragment_id.in_(id_chunk))
                            .returning(
                                self.table.c.fragment_id,
                                self.table.c.document_id,
                                self.table.c.extraction_id,
                                self.table.c.text,
                            )
                        )
                        result = sess.execute(delete_stmt)
                        for row in result:
                            fragment_id = str(row[0])
                            deleted_records[fragment_id] = {
                                "fragment_id": fragment_id,
                                "document_id": str(row[1]),
                                "extraction_id": str(row[2]),
                                "text": row[3],
                            }

                if filters:
                    meta_filter = self.build_filters(filters)
                    delete_stmt = (
                        delete(self.table)
                        .where(meta_filter)
                        .returning(
                            self.table.c.fragment_id,
                            self.table.c.document_id,
                            self.table.c.extraction_id,
                            self.table.c.text,
                        )
                    )
                    result = sess.execute(delete_stmt)
                    for row in result:
                        fragment_id = str(row[0])
                        deleted_records[fragment_id] = {
                            "fragment_id": fragment_id,
                            "document_id": str(row[1]),
                            "extraction_id": str(row[2]),
                            "text": row[3],
                        }
        return deleted_records

    def __getitem__(self, items):
        """
        Fetches a vector from the collection by its identifier.

        Args:
            items (str): The identifier of the vector.

        Returns:
            Record: The fetched vector.
        """
        if not isinstance(items, str):
            raise ArgError("items must be a string id")

        row = self.fetch([items])

        if row == []:
            raise KeyError("no item found with requested id")
        return row[0]

    def query(
        self,
        vector: list[float],
        search_settings: VectorSearchSettings,
    ) -> Union[List[Record], List[str]]:
        """
        Executes a similarity search in the collection.

        The return type is dependent on arguments *include_value* and *include_metadata*

        Args:
            data (list[float]): The vector to use as the query.
            search_settings (VectorSearchSettings): The search settings to use.

        Returns:
            Union[List[Record], List[str]]: The result of the similarity search.
        """

        try:
            imeasure_obj = IndexMeasure(search_settings.index_measure)
        except ValueError:
            raise ArgError("Invalid index measure")

        if not self.is_indexed_for_measure(imeasure_obj):
            warnings.warn(
                UserWarning(
                    f"Query does not have a covering index for {imeasure_obj}. See Collection.create_index"
                )
            )

        distance_lambda = INDEX_MEASURE_TO_SQLA_ACC.get(imeasure_obj)
        if distance_lambda is None:
            # unreachable
            raise ArgError("invalid distance_measure")  # pragma: no cover

        distance_clause = distance_lambda(self.table.c.vec)(vector)

        cols = [
            self.table.c.fragment_id,
            self.table.c.extraction_id,
            self.table.c.document_id,
            self.table.c.user_id,
            self.table.c.group_ids,
            self.table.c.text,
        ]
        if search_settings.include_values:
            cols.append(distance_clause)

        if search_settings.include_metadatas:
            cols.append(self.table.c.metadata)

        stmt = select(*cols)

        # if filters:
        stmt = stmt.filter(self.build_filters(search_settings.filters))  # type: ignore

        stmt = stmt.order_by(distance_clause)
        stmt = stmt.limit(search_settings.search_limit)

        with self.client.Session() as sess:
            with sess.begin():
                # index ignored if greater than n_lists
                sess.execute(
                    text("set local ivfflat.probes = :probes").bindparams(
                        probes=search_settings.probes
                    )
                )
                if self.client._supports_hnsw():
                    sess.execute(
                        text(
                            "set local hnsw.ef_search = :ef_search"
                        ).bindparams(ef_search=search_settings.ef_search)
                    )
                if len(cols) == 1:
                    return [str(x) for x in sess.scalars(stmt).fetchall()]
                return sess.execute(stmt).fetchall() or []

    def full_text_search(
        self, query_text: str, search_settings: VectorSearchSettings
    ) -> List[VectorSearchResult]:

        stemmer = SnowballStemmer("english")

        def expand_query(query: str) -> str:
            words = query.split()
            expanded_words = set()
            for word in words:
                stemmed = stemmer.stem(word)
                expanded_words.update([word, stemmed])
                for syn in wordnet.synsets(word):
                    for lemma in syn.lemmas():
                        expanded_words.add(lemma.name())
            return " | ".join(expanded_words)

        expanded_query = expand_query(query_text)
        ts_query = sa.func.to_tsquery("english", expanded_query)

        phrase_rank = (
            sa.func.ts_rank_cd(
                sa.func.to_tsvector("english", self.table.c.text),
                sa.func.phraseto_tsquery("english", query_text),
                32,
            )
            * 2
        )  # Give higher weight to exact phrases

        partial_match_rank = sa.func.ts_rank_cd(
            sa.func.to_tsvector("english", self.table.c.text),
            sa.func.to_tsquery(
                "english",
                " & ".join(f"{word}:*" for word in query_text.split()),
            ),
            32,
        )

        combined_rank = (
            phrase_rank
            + partial_match_rank
            + sa.func.ts_rank_cd(
                sa.func.to_tsvector("english", self.table.c.text), ts_query, 32
            )
            * (sa.func.similarity(self.table.c.text, query_text) + 0.1)
        ).label("rank")

        stmt = (
            sa.select(
                self.table.c.fragment_id,
                self.table.c.extraction_id,
                self.table.c.document_id,
                self.table.c.user_id,
                self.table.c.group_ids,
                self.table.c.text,
                self.table.c.metadata,
                combined_rank,
            )
            .where(
                sa.and_(
                    sa.or_(
                        self.table.c.fts.op("@@")(ts_query),
                        sa.func.similarity(self.table.c.text, query_text)
                        > 0.1,
                        self.table.c.fts.op("@@")(
                            sa.func.phraseto_tsquery("english", query_text)
                        ),
                        self.table.c.fts.op("@@")(
                            sa.func.to_tsquery(
                                "english",
                                " & ".join(
                                    f"{word}:*" for word in query_text.split()
                                ),
                            )
                        ),
                    ),
                    self.build_filters(search_settings.filters),
                )
            )
            .order_by(sa.desc("rank"))
            .limit(search_settings.hybrid_search_settings.full_text_limit)
        )

        with self.client.Session() as sess:
            results = sess.execute(stmt).fetchall()

        return [
            VectorSearchResult(
                fragment_id=str(r.fragment_id),
                extraction_id=str(r.extraction_id),
                document_id=str(r.document_id),
                user_id=str(r.user_id),
                group_ids=r.group_ids,
                text=r.text,
                score=float(r.rank),
                metadata=r.metadata,
            )
            for r in results
        ]

    def build_filters(self, filters: Dict):
        """
        PUBLIC

        Builds filters for SQL query based on provided dictionary.

        Args:
            table: The SQLAlchemy table object.
            filters (Dict): The dictionary specifying filter conditions.

        Raises:
            FilterError: If filter conditions are not correctly formatted.

        Returns:
            The filter clause for the SQL query.
        """

        if not isinstance(filters, dict):
            raise FilterError("filters must be a dict")

        def parse_condition(key, value):
            if key in Collection.COLUMN_VARS:
                # Handle column-based filters
                column = getattr(self.table.c, key)
                if isinstance(value, dict):
                    op, clause = next(iter(value.items()))
                    if op == "$eq":
                        return column == clause
                    elif op == "$ne":
                        return column != clause
                    elif op == "$in":
                        return column.in_(clause)
                    elif op == "$nin":
                        return ~column.in_(clause)
                    elif op == "$overlap":
                        return column.overlap(clause)
                    elif op == "$contains":
                        return column.contains(clause)
                    elif op == "$any":
                        if key == "group_ids":
                            # Use ANY for UUID array comparison
                            return func.array_to_string(column, ",").like(
                                f"%{clause}%"
                            )
                        # New operator for checking if any element in the array matches
                        return column.any(clause)
                    else:
                        raise FilterError(
                            f"Unsupported operator for column {key}: {op}"
                        )
                else:
                    return column == value
            else:
                # Handle JSON-based filters
                json_col = self.table.c.metadata
                if key.startswith("metadata."):
                    key.split("metadata.")[1]
                if isinstance(value, dict):
                    if len(value) > 1:
                        raise FilterError("only one operator permitted")
                    operator, clause = next(iter(value.items()))
                    if operator not in (
                        "$eq",
                        "$ne",
                        "$lt",
                        "$lte",
                        "$gt",
                        "$gte",
                        "$in",
                        "$contains",
                    ):
                        raise FilterError("unknown operator")

                    if operator == "$eq" and not hasattr(clause, "__len__"):
                        contains_value = cast({key: clause}, postgresql.JSONB)
                        return json_col.op("@>")(contains_value)

                    if operator == "$in":
                        if not isinstance(clause, list):
                            raise FilterError(
                                "argument to $in filter must be a list"
                            )
                        for elem in clause:
                            if not isinstance(elem, (int, str, float)):
                                raise FilterError(
                                    "argument to $in filter must be a list of scalars"
                                )
                        contains_value = [
                            cast(elem, postgresql.JSONB) for elem in clause
                        ]
                        return json_col.op("->")(key).in_(contains_value)

                    matches_value = cast(clause, postgresql.JSONB)

                    if operator == "$contains":
                        if not isinstance(clause, (int, str, float)):
                            raise FilterError(
                                "argument to $contains filter must be a scalar"
                            )
                        return and_(
                            json_col.op("->")(key).contains(matches_value),
                            func.jsonb_typeof(json_col.op("->")(key))
                            == "array",
                        )

                    return {
                        "$eq": json_col.op("->")(key) == matches_value,
                        "$ne": json_col.op("->")(key) != matches_value,
                        "$lt": json_col.op("->")(key) < matches_value,
                        "$lte": json_col.op("->")(key) <= matches_value,
                        "$gt": json_col.op("->")(key) > matches_value,
                        "$gte": json_col.op("->")(key) >= matches_value,
                    }[operator]
                else:
                    contains_value = cast({key: value}, postgresql.JSONB)
                    return json_col.op("@>")(contains_value)

        def parse_filter(filter_dict):
            conditions = []
            for key, value in filter_dict.items():
                if key == "$and":
                    conditions.append(and_(*[parse_filter(f) for f in value]))
                elif key == "$or":
                    conditions.append(or_(*[parse_filter(f) for f in value]))
                else:
                    conditions.append(parse_condition(key, value))
            return and_(*conditions)

        return parse_filter(filters)

    @classmethod
    def _list_collections(cls, client: "Client") -> List["Collection"]:
        """
        PRIVATE

        Retrieves all collections from the database.

        Args:
            client (Client): The database client.

        Returns:
            List[Collection]: A list of all existing collections.
        """

        query = text(
            """
        select
            relname as table_name,
            atttypmod as embedding_dim
        from
            pg_class pc
            join pg_attribute pa
                on pc.oid = pa.attrelid
        where
            pc.relnamespace = 'vecs'::regnamespace
            and pc.relkind = 'r'
            and pa.attname = 'vec'
            and not pc.relname ^@ '_'
        """
        )
        xc = []
        with client.Session() as sess:
            for name, dimension in sess.execute(query):
                existing_collection = cls(name, dimension, client)
                xc.append(existing_collection)
        return xc

    @classmethod
    def _does_collection_exist(cls, client: "Client", name: str) -> bool:
        """
        PRIVATE

        Checks if a collection with a given name exists within the database

        Args:
            client (Client): The database client.
            name (str): The name of the collection

        Returns:
            Exists: Whether the collection exists or not
        """

        try:
            client.get_collection(name)
            return True
        except CollectionNotFound:
            return False

    @property
    def index(self) -> Optional[str]:
        """
        PRIVATE

        Note:
            The `index` property is private and expected to undergo refactoring.
            Do not rely on it's output.

        Retrieves the SQL name of the collection's vector index, if it exists.

        Returns:
            Optional[str]: The name of the index, or None if no index exists.
        """

        if self._index is None:
            query = text(
                """
            select
                relname as table_name
            from
                pg_class pc
            where
                pc.relnamespace = 'vecs'::regnamespace
                and relname ilike 'ix_vector%'
                and pc.relkind = 'i'
            """
            )
            with self.client.Session() as sess:
                ix_name = sess.execute(query).scalar()
            self._index = ix_name
        return self._index

    def is_indexed_for_measure(self, measure: IndexMeasure):
        """
        Checks if the collection is indexed for a specific measure.

        Args:
            measure (IndexMeasure): The measure to check for.

        Returns:
            bool: True if the collection is indexed for the measure, False otherwise.
        """

        index_name = self.index
        if index_name is None:
            return False

        ops = INDEX_MEASURE_TO_OPS.get(measure)
        if ops is None:
            return False

        if ops in index_name:
            return True

        return False

    def create_index(
        self,
        measure: IndexMeasure = IndexMeasure.cosine_distance,
        method: IndexMethod = IndexMethod.auto,
        index_arguments: Optional[
            Union[IndexArgsIVFFlat, IndexArgsHNSW]
        ] = None,
        replace=True,
    ) -> None:
        """
        Creates an index for the collection.

        Note:
            When `vecs` creates an index on a pgvector column in PostgreSQL, it uses a multi-step
            process that enables performant indexes to be built for large collections with low end
            database hardware.

            Those steps are:

            - Creates a new table with a different name
            - Randomly selects records from the existing table
            - Inserts the random records from the existing table into the new table
            - Creates the requested vector index on the new table
            - Upserts all data from the existing table into the new table
            - Drops the existing table
            - Renames the new table to the existing tables name

            If you create dependencies (like views) on the table that underpins
            a `vecs.Collection` the `create_index` step may require you to drop those dependencies before
            it will succeed.

        Args:
            measure (IndexMeasure, optional): The measure to index for. Defaults to 'cosine_distance'.
            method (IndexMethod, optional): The indexing method to use. Defaults to 'auto'.
            index_arguments: (IndexArgsIVFFlat | IndexArgsHNSW, optional): Index type specific arguments
            replace (bool, optional): Whether to replace the existing index. Defaults to True.

        Raises:
            ArgError: If an invalid index method is used, or if *replace* is False and an index already exists.
        """

        if method not in (
            IndexMethod.ivfflat,
            IndexMethod.hnsw,
            IndexMethod.auto,
        ):
            raise ArgError("invalid index method")

        if index_arguments:
            # Disallow case where user submits index arguments but uses the
            # IndexMethod.auto index (index build arguments should only be
            # used with a specific index)
            if method == IndexMethod.auto:
                raise ArgError(
                    "Index build parameters are not allowed when using the IndexMethod.auto index."
                )
            # Disallow case where user specifies one index type but submits
            # index build arguments for the other index type
            if (
                isinstance(index_arguments, IndexArgsHNSW)
                and method != IndexMethod.hnsw
            ) or (
                isinstance(index_arguments, IndexArgsIVFFlat)
                and method != IndexMethod.ivfflat
            ):
                raise ArgError(
                    f"{index_arguments.__class__.__name__} build parameters were supplied but {method} index was specified."
                )

        if method == IndexMethod.auto:
            if self.client._supports_hnsw():
                method = IndexMethod.hnsw
            else:
                method = IndexMethod.ivfflat

        if method == IndexMethod.hnsw and not self.client._supports_hnsw():
            raise ArgError(
                "HNSW Unavailable. Upgrade your pgvector installation to > 0.5.0 to enable HNSW support"
            )

        ops = INDEX_MEASURE_TO_OPS.get(measure)
        if ops is None:
            raise ArgError("Unknown index measure")

        unique_string = str(uuid4()).replace("-", "_")[0:7]

        with self.client.Session() as sess:
            with sess.begin():
                if self.index is not None:
                    if replace:
                        sess.execute(text(f'drop index vecs."{self.index}";'))
                        self._index = None
                    else:
                        raise ArgError(
                            "replace is set to False but an index exists"
                        )

                if method == IndexMethod.ivfflat:
                    if not index_arguments:
                        n_records: int = sess.execute(func.count(self.table.c.id)).scalar()  # type: ignore

                        n_lists = (
                            int(max(n_records / 1000, 30))
                            if n_records < 1_000_000
                            else int(math.sqrt(n_records))
                        )
                    else:
                        # The following mypy error is ignored because mypy
                        # complains that `index_arguments` is typed as a union
                        # of IndexArgsIVFFlat and IndexArgsHNSW types,
                        # which both don't necessarily contain the `n_lists`
                        # parameter, however we have validated that the
                        # correct type is being used above.
                        n_lists = index_arguments.n_lists  # type: ignore

                    sess.execute(
                        text(
                            f"""
                            create index ix_{ops}_ivfflat_nl{n_lists}_{unique_string}
                              on vecs."{self.table.name}"
                              using ivfflat (vec {ops}) with (lists={n_lists})
                            """
                        )
                    )

                if method == IndexMethod.hnsw:
                    if not index_arguments:
                        index_arguments = IndexArgsHNSW()

                    # See above for explanation of why the following lines
                    # are ignored
                    m = index_arguments.m  # type: ignore
                    ef_construction = index_arguments.ef_construction  # type: ignore

                    sess.execute(
                        text(
                            f"""
                            create index ix_{ops}_hnsw_m{m}_efc{ef_construction}_{unique_string}
                              on vecs."{self.table.name}"
                              using hnsw (vec {ops}) WITH (m={m}, ef_construction={ef_construction});
                            """
                        )
                    )

        return None


def _build_table(name: str, meta: MetaData, dimension: int) -> Table:
    table = Table(
        name,
        meta,
        Column("fragment_id", postgresql.UUID, primary_key=True),
        Column("extraction_id", postgresql.UUID, nullable=False),
        Column("document_id", postgresql.UUID, nullable=False),
        Column("user_id", postgresql.UUID, nullable=False),
        Column(
            "group_ids", postgresql.ARRAY(postgresql.UUID), server_default="{}"
        ),
        Column("vec", Vector(dimension), nullable=False),
        Column("text", postgresql.TEXT, nullable=True),
        Column(
            "fts",
            postgresql.TSVECTOR,
            nullable=False,
            server_default=text("to_tsvector('english'::regconfig, '')"),
        ),
        Column(
            "metadata",
            postgresql.JSONB,
            server_default=text("'{}'::jsonb"),
            nullable=False,
        ),
        extend_existing=True,
    )

    # # Add GIN index for full-text search and trigram similarity
    Index(
        f"idx_{name}_fts_trgm",
        table.c.fts,
        table.c.text,
        postgresql_using="gin",
        postgresql_ops={
            "text": "gin_trgm_ops"
        },  # alternative,  gin_tsvector_ops
    )
    return table



from typing import TYPE_CHECKING, Optional

from core.base import RunLoggingSingleton, RunManager
from core.base.abstractions.base import AsyncSyncMeta, syncable

from .abstractions import R2RAgents, R2RPipelines, R2RProviders
from .services.auth_service import AuthService
from .services.ingestion_service import IngestionService
from .services.management_service import ManagementService
from .services.restructure_service import RestructureService
from .services.retrieval_service import RetrievalService

if TYPE_CHECKING:
    from .assembly.config import R2RConfig


class R2REngine(metaclass=AsyncSyncMeta):
    def __init__(
        self,
        config: "R2RConfig",
        providers: R2RProviders,
        pipelines: R2RPipelines,
        agents: R2RAgents,
        run_manager: Optional[RunManager] = None,
    ):
        logging_connection = RunLoggingSingleton()
        run_manager = run_manager or RunManager(logging_connection)

        self.config = config
        self.providers = providers
        self.pipelines = pipelines
        self.agents = agents
        self.logging_connection = RunLoggingSingleton()
        self.run_manager = run_manager

        self.ingestion_service = IngestionService(
            config,
            providers,
            pipelines,
            agents,
            run_manager,
            logging_connection,
        )

        self.restructure_service = RestructureService(
            config,
            providers,
            pipelines,
            agents,
            run_manager,
            logging_connection,
        )

        self.retrieval_service = RetrievalService(
            config,
            providers,
            pipelines,
            agents,
            run_manager,
            logging_connection,
        )
        self.management_service = ManagementService(
            config,
            providers,
            pipelines,
            agents,
            run_manager,
            logging_connection,
        )

        self.auth_service = AuthService(
            config,
            providers,
            pipelines,
            agents,
            run_manager,
            logging_connection,
        )

    @syncable
    async def aingest_documents(self, *args, **kwargs):
        return await self.ingestion_service.ingest_documents(*args, **kwargs)

    @syncable
    async def aingest_files(self, *args, **kwargs):
        return await self.ingestion_service.ingest_files(*args, **kwargs)

    @syncable
    async def aupdate_files(self, *args, **kwargs):
        return await self.ingestion_service.update_files(*args, **kwargs)

    @syncable
    async def aenrich_graph(self, *args, **kwargs):
        return await self.restructure_service.enrich_graph(*args, **kwargs)

    @syncable
    async def asearch(self, *args, **kwargs):
        return await self.retrieval_service.search(*args, **kwargs)

    @syncable
    async def arag(self, *args, **kwargs):
        return await self.retrieval_service.rag(*args, **kwargs)

    @syncable
    async def arag_agent(self, *args, **kwargs):
        return await self.retrieval_service.agent(*args, **kwargs)

    @syncable
    async def aupdate_prompt(self, *args, **kwargs):
        return await self.management_service.update_prompt(*args, **kwargs)

    @syncable
    async def alogs(self, *args, **kwargs):
        return await self.management_service.alogs(*args, **kwargs)

    @syncable
    async def aanalytics(self, *args, **kwargs):
        return await self.management_service.aanalytics(*args, **kwargs)

    @syncable
    async def aapp_settings(self, *args, **kwargs):
        return await self.management_service.aapp_settings(*args, **kwargs)

    @syncable
    async def ascore_completion(self, *args, **kwargs):
        return await self.management_service.ascore_completion(*args, **kwargs)

    @syncable
    async def ausers_overview(self, *args, **kwargs):
        return await self.management_service.ausers_overview(*args, **kwargs)

    @syncable
    async def adelete(self, *args, **kwargs):
        return await self.management_service.delete(*args, **kwargs)

    @syncable
    async def adocuments_overview(self, *args, **kwargs):
        return await self.management_service.adocuments_overview(
            *args, **kwargs
        )

    @syncable
    async def ainspect_knowledge_graph(self, *args, **kwargs):
        return await self.management_service.inspect_knowledge_graph(
            *args, **kwargs
        )

    @syncable
    async def adocument_chunks(self, *args, **kwargs):
        return await self.management_service.document_chunks(*args, **kwargs)

    @syncable
    async def aregister(self, *args, **kwargs):
        return await self.auth_service.register(*args, **kwargs)

    @syncable
    async def averify_email(self, *args, **kwargs):
        return await self.auth_service.verify_email(*args, **kwargs)

    @syncable
    async def alogin(self, *args, **kwargs):
        return await self.auth_service.login(*args, **kwargs)

    @syncable
    async def auser(self, *args, **kwargs):
        return await self.auth_service.user(*args, **kwargs)

    @syncable
    async def aupdate_user(self, *args, **kwargs):
        return await self.auth_service.update_user(*args, **kwargs)

    @syncable
    async def arefresh_access_token(self, *args, **kwargs):
        return await self.auth_service.refresh_access_token(*args, **kwargs)

    @syncable
    async def achange_password(self, *args, **kwargs):
        return await self.auth_service.change_password(*args, **kwargs)

    @syncable
    async def arequest_password_reset(self, *args, **kwargs):
        return await self.auth_service.request_password_reset(*args, **kwargs)

    @syncable
    async def aconfirm_password_reset(self, *args, **kwargs):
        return await self.auth_service.confirm_password_reset(*args, **kwargs)

    @syncable
    async def alogout(self, *args, **kwargs):
        return await self.auth_service.logout(*args, **kwargs)

    @syncable
    async def adelete_user(self, *args, **kwargs):
        return await self.auth_service.delete_user(*args, **kwargs)

    @syncable
    async def aclean_expired_blacklisted_tokens(self, *args, **kwargs):
        return await self.auth_service.clean_expired_blacklisted_tokens(
            *args, **kwargs
        )

    @syncable
    async def acreate_group(self, *args, **kwargs):
        return await self.management_service.acreate_group(*args, **kwargs)

    @syncable
    async def aget_group(self, *args, **kwargs):
        return await self.management_service.aget_group(*args, **kwargs)

    @syncable
    async def aupdate_group(self, *args, **kwargs):
        return await self.management_service.aupdate_group(*args, **kwargs)

    @syncable
    async def adelete_group(self, *args, **kwargs):
        return await self.management_service.adelete_group(*args, **kwargs)

    @syncable
    async def alist_groups(self, *args, **kwargs):
        return await self.management_service.alist_groups(*args, **kwargs)

    @syncable
    async def aadd_user_to_group(self, *args, **kwargs):
        return await self.management_service.aadd_user_to_group(
            *args, **kwargs
        )

    @syncable
    async def aremove_user_from_group(self, *args, **kwargs):
        return await self.management_service.aremove_user_from_group(
            *args, **kwargs
        )

    @syncable
    async def aget_users_in_group(self, *args, **kwargs):
        return await self.management_service.aget_users_in_group(
            *args, **kwargs
        )

    @syncable
    async def aget_groups_for_user(self, *args, **kwargs):
        return await self.management_service.aget_groups_for_user(
            *args, **kwargs
        )

    @syncable
    async def aassign_document_to_group(self, *args, **kwargs):
        return await self.management_service.aassign_document_to_group(
            *args, **kwargs
        )

    @syncable
    async def agroups_overview(self, *args, **kwargs):
        return await self.management_service.agroups_overview(*args, **kwargs)

    @syncable
    async def adocuments_in_group(self, *args, **kwargs):
        return await self.management_service.adocuments_in_group(
            *args, **kwargs
        )

    @syncable
    async def adocument_groups(self, *args, **kwargs):
        return await self.management_service.adocument_groups(*args, **kwargs)

    @syncable
    async def aremove_document_from_group(self, *args, **kwargs):
        return await self.management_service.aremove_document_from_group(
            *args, **kwargs
        )



import base64
import os
from io import BytesIO
from typing import AsyncGenerator

from PIL import Image

from core.base.abstractions.document import DataType
from core.base.parsers.base_parser import AsyncParser
from core.parsers.media.openai_helpers import process_frame_with_openai


class ImageParser(AsyncParser[DataType]):
    """A parser for image data."""

    def __init__(
        self,
        model: str = "gpt-4o",
        max_tokens: int = 2_048,
        api_base: str = "https://api.openai.com/v1/chat/completions",
        max_image_size: int = 1 * 1024 * 1024,  # 4MB limit
    ):
        self.model = model
        self.max_tokens = max_tokens
        self.openai_api_key = os.environ.get("OPENAI_API_KEY")
        if not self.openai_api_key:
            raise ValueError(
                "Error, environment variable `OPENAI_API_KEY` is required to run `ImageParser`."
            )
        self.api_base = api_base
        self.max_image_size = max_image_size

    def _resize_image(self, image_data: bytes, compression_ratio) -> bytes:
        img = Image.open(BytesIO(image_data))
        img_byte_arr = BytesIO()
        img.save(
            img_byte_arr, format="JPEG", quality=int(100 * compression_ratio)
        )
        return img_byte_arr.getvalue()

    async def ingest(self, data: DataType) -> AsyncGenerator[str, None]:
        """Ingest image data and yield a description."""
        if isinstance(data, bytes):
            # Resize the image if it's too large
            if len(data) > self.max_image_size:
                data = self._resize_image(
                    data, float(self.max_image_size) / len(data)
                )

            # Encode to base64
            data = base64.b64encode(data).decode("utf-8")

        yield process_frame_with_openai(
            data,
            self.openai_api_key,
            self.model,
            self.max_tokens,
            self.api_base,
        )



import json
import logging
from pathlib import Path
from typing import List, Optional
from uuid import UUID

import yaml
from fastapi import Depends, File, Form, UploadFile
from pydantic import Json

from core.base import ChunkingConfig, R2RException
from core.base.api.models.ingestion.responses import WrappedIngestionResponse
from core.base.utils import generate_user_document_id

from ....assembly.factory import R2RProviderFactory
from ....engine import R2REngine
from ..base_router import BaseRouter, RunType

logger = logging.getLogger(__name__)


class IngestionRouter(BaseRouter):
    def __init__(
        self, engine: R2REngine, run_type: RunType = RunType.INGESTION
    ):
        super().__init__(engine, run_type)
        self.openapi_extras = self.load_openapi_extras()
        self.setup_routes()

    def load_openapi_extras(self):
        yaml_path = Path(__file__).parent / "ingestion_router_openapi.yml"
        with open(yaml_path, "r") as yaml_file:
            yaml_content = yaml.safe_load(yaml_file)
        return yaml_content

    def setup_routes(self):
        # Note, we use the following verbose input parameters because FastAPI struggles to handle `File` input and `Body` inputs
        # at the same time. Therefore, we must ues `Form` inputs for the metadata, document_ids
        ingest_files_extras = self.openapi_extras.get("ingest_files", {})
        ingest_files_descriptions = ingest_files_extras.get(
            "input_descriptions", {}
        )

        @self.router.post(
            "/ingest_files",
            openapi_extra=ingest_files_extras.get("openapi_extra"),
        )
        @self.base_endpoint
        async def ingest_files_app(
            files: List[UploadFile] = File(
                ..., description=ingest_files_descriptions.get("files")
            ),
            document_ids: Optional[Json[list[UUID]]] = Form(
                None,
                description=ingest_files_descriptions.get("document_ids"),
            ),
            metadatas: Optional[Json[list[dict]]] = Form(
                None, description=ingest_files_descriptions.get("metadatas")
            ),
            chunking_settings: Optional[Json[ChunkingConfig]] = Form(
                None,
                description=ingest_files_descriptions.get("chunking_settings"),
            ),
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ) -> WrappedIngestionResponse:
            """
            Ingest files into the system.

            This endpoint supports multipart/form-data requests, enabling you to ingest files and their associated metadatas into R2R.

            A valid user authentication token is required to access this endpoint, as regular users can only ingest files for their own access. More expansive group permissioning is under development.
            """
            chunking_provider = None
            if chunking_settings:
                chunking_settings.validate()
                chunking_provider = (
                    R2RProviderFactory.create_chunking_provider(
                        chunking_settings
                    )
                )
            else:
                logger.info(
                    "No chunking config override provided. Using default."
                )

            # Check if the user is a superuser
            is_superuser = auth_user and auth_user.is_superuser

            # Handle user management logic at the request level
            if not auth_user:
                for metadata in metadatas or []:
                    if "user_id" in metadata and (
                        not is_superuser
                        and metadata["user_id"] != str(auth_user.id)
                    ):
                        raise R2RException(
                            status_code=403,
                            message="Non-superusers cannot set user_id in metadata.",
                        )

                # If user is not a superuser, set user_id in metadata
                metadata["user_id"] = str(auth_user.id)

            ingestion_result = await self.engine.aingest_files(
                files=files,
                metadatas=metadatas,
                document_ids=document_ids,
                user=auth_user,
                chunking_provider=chunking_provider,
            )

            # If superuser, assign documents to groups
            if is_superuser:
                for idx, metadata in enumerate(metadatas or []):
                    if "group_ids" in metadata:
                        document_id = ingestion_result["processed_documents"][
                            idx
                        ]
                        for group_id in metadata["group_ids"]:
                            await self.engine.management_service.aassign_document_to_group(
                                document_id, group_id
                            )

            return ingestion_result

        update_files_extras = self.openapi_extras.get("update_files", {})
        update_files_descriptions = update_files_extras.get(
            "input_descriptions", {}
        )

        @self.router.post(
            "/update_files",
            openapi_extra=update_files_extras.get("openapi_extra"),
        )
        @self.base_endpoint
        async def update_files_app(
            files: List[UploadFile] = File(
                ..., description=update_files_descriptions.get("files")
            ),
            document_ids: Optional[Json[list[UUID]]] = Form(
                None,
                description=ingest_files_descriptions.get("document_ids"),
            ),
            metadatas: Optional[Json[list[dict]]] = Form(
                None, description=ingest_files_descriptions.get("metadatas")
            ),
            chunking_settings: Optional[Json[ChunkingConfig]] = Form(
                None,
                description=ingest_files_descriptions.get("chunking_settings"),
            ),
            auth_user=Depends(self.engine.providers.auth.auth_wrapper),
        ) -> WrappedIngestionResponse:
            """
            Update existing files in the system.

            This endpoint supports multipart/form-data requests, enabling you to update files and their associated metadatas into R2R.




            A valid user authentication token is required to access this endpoint, as regular users can only update their own files. More expansive group permissioning is under development.
            """

            chunking_provider = None
            if chunking_settings:
                config = ChunkingConfig(**chunking_settings)
                chunking_provider = (
                    R2RProviderFactory.create_chunking_provider(config)
                )

            return await self.engine.aupdate_files(
                files=files,
                metadatas=metadatas,
                document_ids=document_ids,
                chunking_provider=chunking_provider,
                user=auth_user,
            )



import fire

from r2r import R2RBuilder, R2RConfig, R2RPipeFactoryWithMultiSearch


def main(task_prompt_name="hyde"):
    # Load the default configuration file
    config = R2RConfig.from_toml()

    app = (
        R2RBuilder(config)
        .with_pipe_factory(R2RPipeFactoryWithMultiSearch)
        .build(
            # Add optional override arguments which propagate to the pipe factory
            task_prompt_name=task_prompt_name,
        )
    )
    app.serve()


if __name__ == "__main__":
    fire.Fire(main)



from .auth import AuthConfig, AuthProvider
from .base import Provider, ProviderConfig
from .chunking import ChunkingConfig, ChunkingProvider, Method
from .crypto import CryptoConfig, CryptoProvider
from .database import (
    DatabaseConfig,
    DatabaseProvider,
    RelationalDBProvider,
    VectorDBProvider,
)
from .embedding import EmbeddingConfig, EmbeddingProvider
from .kg import KGConfig, KGProvider
from .llm import CompletionConfig, CompletionProvider
from .parsing import OverrideParser, ParsingConfig, ParsingProvider
from .prompt import PromptConfig, PromptProvider

__all__ = [
    # Base provider classes
    "Provider",
    "ProviderConfig",
    # Auth provider
    "AuthConfig",
    "AuthProvider",
    # Chunking provider
    "ChunkingConfig",
    "ChunkingProvider",
    "Method",
    # Crypto provider
    "CryptoConfig",
    "CryptoProvider",
    # Database providers
    "DatabaseConfig",
    "DatabaseProvider",
    "RelationalDBProvider",
    "VectorDBProvider",
    # Embedding provider
    "EmbeddingConfig",
    "EmbeddingProvider",
    # Knowledge Graph provider
    "KGConfig",
    "KGProvider",
    # LLM provider
    "CompletionConfig",
    "CompletionProvider",
    # Parsing provider
    "ParsingConfig",
    "ParsingProvider",
    "OverrideParser",
    # Prompt provider
    "PromptConfig",
    "PromptProvider",
]



from sdk import *

__all__ = [
    # R2R SDK
    "R2RAsyncClient",
    "R2RClient",
]

try:
    import core
    from core import *

    __all__ += core.__all__
except ImportError:
    # Core dependencies not installed
    pass



from typing import Any, List, TypeVar
from uuid import UUID

from pydantic import BaseModel, Field

from core.base.api.models.base import ResultsWrapper

T = TypeVar("T")


class ProcessedDocument(BaseModel):
    id: UUID
    title: str


class FailedDocument(BaseModel):
    document_id: UUID
    result: Any


class IngestionResponse(BaseModel):
    processed_documents: List[ProcessedDocument] = Field(
        ...,
        description="List of successfully processed documents",
        example=[
            {
                "id": "123e4567-e89b-12d3-a456-426614174000",
                "title": "Document 1",
            },
            {
                "id": "223e4567-e89b-12d3-a456-426614174000",
                "title": "Document 2",
            },
        ],
    )
    failed_documents: List[FailedDocument] = Field(
        ...,
        description="List of documents that failed to process",
        example=[
            {
                "document_id": "323e4567-e89b-12d3-a456-426614174000",
                "result": "Error: Invalid format",
            }
        ],
    )
    skipped_documents: List[UUID] = Field(
        ...,
        description="List of document IDs that were skipped during processing",
        example=["423e4567-e89b-12d3-a456-426614174000"],
    )

    class Config:
        json_schema_extra = {
            "example": {
                "processed_documents": [
                    {
                        "id": "123e4567-e89b-12d3-a456-426614174000",
                        "title": "Document 1",
                    },
                    {
                        "id": "223e4567-e89b-12d3-a456-426614174000",
                        "title": "Document 2",
                    },
                ],
                "failed_documents": [
                    {
                        "document_id": "323e4567-e89b-12d3-a456-426614174000",
                        "result": "Error: Invalid format",
                    }
                ],
                "skipped_documents": ["423e4567-e89b-12d3-a456-426614174000"],
            }
        }


# Create wrapped version of the response
WrappedIngestionResponse = ResultsWrapper[IngestionResponse]



from datetime import datetime, timedelta
from typing import Optional

from .base import DatabaseMixin, QueryBuilder


class BlacklistedTokensMixin(DatabaseMixin):
    def create_table(self):
        query = f"""
        CREATE TABLE IF NOT EXISTS {self._get_table_name('blacklisted_tokens')} (
            id UUID PRIMARY KEY DEFAULT uuid_generate_v4(),
            token TEXT NOT NULL,
            blacklisted_at TIMESTAMPTZ DEFAULT NOW()
        );
        CREATE INDEX IF NOT EXISTS idx_blacklisted_tokens_{self.collection_name}_token
        ON {self._get_table_name('blacklisted_tokens')} (token);
        CREATE INDEX IF NOT EXISTS idx_blacklisted_tokens_{self.collection_name}_blacklisted_at
        ON {self._get_table_name('blacklisted_tokens')} (blacklisted_at);
        """
        self.execute_query(query)

    def blacklist_token(self, token: str, current_time: datetime = None):
        if current_time is None:
            current_time = datetime.utcnow()

        query, params = (
            QueryBuilder(self._get_table_name("blacklisted_tokens"))
            .insert({"token": token, "blacklisted_at": current_time})
            .build()
        )
        self.execute_query(query, params)

    def is_token_blacklisted(self, token: str) -> bool:
        query, params = (
            QueryBuilder(self._get_table_name("blacklisted_tokens"))
            .select(["1"])
            .where("token = :token", token=token)
            .limit(1)
            .build()
        )
        result = self.execute_query(query, params)
        return bool(result.fetchone())

    def clean_expired_blacklisted_tokens(
        self,
        max_age_hours: int = 7 * 24,
        current_time: Optional[datetime] = None,
    ):
        if current_time is None:
            current_time = datetime.utcnow()
        expiry_time = current_time - timedelta(hours=max_age_hours)

        query, params = (
            QueryBuilder(self._get_table_name("blacklisted_tokens"))
            .delete()
            .where("blacklisted_at < :expiry_time", expiry_time=expiry_time)
            .build()
        )
        self.execute_query(query, params)



from typing import Any, Dict

from pydantic import BaseModel

from core.base.api.models.base import ResultsWrapper


class KGEnrichmentResponse(BaseModel):
    enriched_content: Dict[str, Any]


WrappedKGEnrichmentResponse = ResultsWrapper[KGEnrichmentResponse]
